"""
test_centers.py
"""
def test_calc_radii():
	import numpy as np
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	df = pd.read_csv("dash.csv").query('epitope == "PA"').reset_index(drop = True)
	tr = TCRrep(cell_df = df.copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
	            compute_distances = True)

	df = pd.read_csv("dash.csv").query('epitope != "PA"').reset_index(drop = True)
	tr_bkgd = TCRrep(cell_df = df.copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
               compute_distances = False)

	from tcrdist.centers import calc_radii
	radii, thresholds, ecdfs = calc_radii(tr = tr, tr_bkgd = tr_bkgd, chain = 'beta', ctrl_bkgd = 10**-3, use_sparse = False, max_radius=50)
	from tcrdist.public import _neighbors_variable_radius
	# Compute neighbors <= variable radius in the background set and the foreground set
	neighbors     = _neighbors_variable_radius(pwmat = tr.pw_beta , radius_list = radii)
	background_neighbors = _neighbors_variable_radius(pwmat = tr.rw_beta , radius_list = radii)
	tr.clone_df['radius']               = radii
	tr.clone_df['neighbors']            = neighbors
	tr.clone_df['background_neighbors'] = background_neighbors
	tr.clone_df['nsubject']             = tr.clone_df['neighbors'].\
			apply(lambda x: tr.clone_df['subject'].iloc[x].nunique())
	tr.clone_df['qpublic']              = tr.clone_df['nsubject'].\
			apply(lambda x: x > 1)
	tr.clone_df


def test_calc_radii_if_big():
	import numpy as np
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	from tcrdist.centers import calc_radii
	from tcrdist.public import _neighbors_sparse_variable_radius
	df = pd.read_csv("dash.csv").query('epitope == "PA"').reset_index(drop = True)
	tr = TCRrep(cell_df = df.copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
	            compute_distances = False)
	# For a large matrix one can use compute_sparse_rect_distances() instead of
	# .compute_distances() for the pairwise square matrix
	tr.cpus = 2
	tr.compute_sparse_rect_distances(df = tr.clone_df, radius=50,chunk_size=100)
	tr.pw_beta = tr.rw_beta.copy()

	# 
	dfb = pd.read_csv("dash.csv").query('epitope != "PA"').reset_index(drop = True)
	tr_bkgd = TCRrep(cell_df = dfb.copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = False)
	
	# Set rw_beta to none as it will be computed between target and background by calc_radii
	tr.rw_beta = None
	from tcrdist.centers import calc_radii
	radii, thresholds, ecdfs = calc_radii(tr = tr, tr_bkgd = tr_bkgd, chain = 'beta', ctrl_bkgd = 10**-3, use_sparse = True, max_radius=50)
	tr.clone_df['radius'] = radii

	from tcrdist.public import _neighbors_sparse_variable_radius
	# Compute neighbors <= variable radius in the background set and the foreground set
	neighbors     = _neighbors_sparse_variable_radius(csrmat = tr.pw_beta , radius_list = tr.clone_df['radius'])
	background_neighbors = _neighbors_sparse_variable_radius(csrmat = tr.rw_beta , radius_list = tr.clone_df['radius'])
	#tr.clone_df['radius']               = radii
	tr.clone_df['neighbors']            = neighbors
	tr.clone_df['background_neighbors'] = background_neighbors
	tr.clone_df['nsubject']             = tr.clone_df['neighbors'].\
			apply(lambda x: tr.clone_df['subject'].iloc[x].nunique())
	tr.clone_df['qpublic']              = tr.clone_df['nsubject'].\
			apply(lambda x: x > 1)
	tr.clone_df



def test_example_with_report():
	# Example that would work with a large bakcgournd
	import numpy as np
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	from tcrdist.background import sample_britanova
	from tcrdist.sample import _default_sampler
	"""A useful background for beta chain"""
	df = pd.read_csv("dash.csv").query('epitope == "PA"').reset_index(drop = True)

	tr = TCRrep(cell_df = df.copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
	            compute_distances = True)
	
	ts = _default_sampler(organism = "mouse", chain = "beta")()
	trb = TCRrep(cell_df = ts.ref_df.rename(columns = {'v_reps' : 'v_b_gene', 'j_reps': 'j_b_gene', 'cdr3': 'cdr3_b_aa'}).copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
	            compute_distances = False)

	tr.cpus = 2
	tr.compute_sparse_rect_distances(df = tr.clone_df, df2 = trb.clone_df, radius=50,chunk_size=100)
	
	from tcrdist.centers import calc_radii
	radii, thresholds, ecdfs = calc_radii(tr = tr, tr_bkgd = trb, chain = 'beta', ctrl_bkgd = 10**-5, use_sparse = True, max_radius=50)

	# Set a maximum radius of 26
	tr.clone_df['radius'] = radii
	tr.clone_df['radius'][tr.clone_df['radius'] > 26] = 26

	# Quick access to publicity
	from tcrdist.public import _neighbors_sparse_variable_radius, _neighbors_variable_radius
	tr.clone_df['neighbors'] = _neighbors_variable_radius(pwmat = tr.pw_beta, radius_list = tr.clone_df['radius'])
	tr.clone_df['background_neighbors'] = _neighbors_sparse_variable_radius(csrmat = tr.rw_beta, radius_list = tr.clone_df['radius'])
	tr.clone_df['nsubject']             = tr.clone_df['neighbors'].\
			apply(lambda x: tr.clone_df['subject'].iloc[x].nunique())
	tr.clone_df['qpublic']              = tr.clone_df['nsubject'].\
			apply(lambda x: x > 1)
	tr.clone_df

	# A Report
	from tcrdist.public import TCRpublic
	tp = TCRpublic(
	tcrrep = tr, 
	output_html_name = "quasi_public_clones.html")
	tp.fixed_radius = False
	rp = tp.report()





def alternative_actually_a_big_case():
	import os
	import numpy as np
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	from tcrdist.neighbors import compute_population_estimate_ecdf
	from tcrdist.ecdf import distance_ecdf
	project_path = os.path.join('tutorial')
	source_path = os.path.join('tcrdist','data','covid19')
	antigen_enriched_background_file = 'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv'
	assert os.path.isfile(os.path.join(source_path, antigen_enriched_background_file))
		# Read file into a Pandas DataFrame <df>
	df = pd.read_csv(os.path.join(source_path, antigen_enriched_background_file))
		# Drop cells without any gene usage information
	df = df.query("v_b_gene.notna() & j_b_gene.notna()")
		# Provide a counts column if non is present
	df['count'] = 1
		# Initialize a TCRrep class, using ONLY columns that are complete and unique define a a clone.
		# Counts of identical 'clones' will be aggregated into a TCRrep.clone_df.
	tr = TCRrep(cell_df = df[['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'subject','count']], 
				organism = "human", 
				chains = ['beta'], 
				compute_distances = True,
				cpus = 6)

	df_bkgd = pd.read_csv(os.path.join(project_path, f"{antigen_enriched_background_file}.olga100K_brit100K_bkgd_2.csv"), sep = ",")
		# Load the background to a TCRrep without computing pairwise distances (i.e., compute_distances = False)
	tr_bkgd = TCRrep(cell_df = df_bkgd, organism = "human", chains = ['beta'], compute_distances = False)

	from tcrdist.centers import calc_radii

	radii = calc_radii(tr = tr, tr_bkgd = tr_bkgd,  ctrl_bkgd = 10**-5, use_sparse = True, max_radius=50, chunk_size=100)
	radii2 = calc_radii(tr = tr, tr_bkgd = tr_bkgd,  ctrl_bkgd = 10**-5, use_sparse = False, max_radius = 50)
	assert radii == radii2

	tr.clone_df['max_radii'] = radii


	






	# # Set to None so it is recomputed as 
	# tr.rw_beta = None
	# radii2, thresholds2, ecdfs2 = calc_radii(tr = tr, tr_bkgd = tr_bkgd, chain = 'beta', ctrl_bkgd = 10**-3, use_sparse = True, max_radius=50, chunk_size = 100)
	# assert len(radii) == tr.clone_df.shape[0]
	# assert np.all(radii == radii2)
	# assert np.all(thresholds == thresholds2)
	# assert np.all(ecdfs  == ecdfs2)
	# from tcrdist.public import _neighbors_sparse_variable_radius
	# background_neighbors2 = _neighbors_sparse_variable_radius(csrmat = tr.rw_beta, radius_list = radii2, maxd= 50)
	# np.all(background_neighbors == background_neighbors2)


















	# # Example with a report
	# import numpy as np
	# import pandas as pd
	# from tcrdist.repertoire import TCRrep
	# df = pd.read_csv("dash.csv").query('epitope == "PA"').reset_index(drop = True)
	# tr = TCRrep(cell_df = df.copy(), 
	#             organism = 'mouse', 
	#             chains = ['beta'], 
	#             db_file = 'alphabeta_gammadelta_db.tsv',
	#             compute_distances = True)

	# df = pd.read_csv("dash.csv").query('epitope != "PA"').reset_index(drop = True)
	# tr_bkgd = TCRrep(cell_df = df.copy(), 
	#             organism = 'mouse', 
	#             chains = ['beta'], 
	#             db_file = 'alphabeta_gammadelta_db.tsv',
	#             compute_distances = False)
	# radii, thresholds, ecdfs = calc_radii(tr = tr, tr_bkgd = tr_bkgd, chain = 'beta', ctrl_bkgd = 10**-5, use_sparse = False, max_radius=50)
	# tr.clone_df['radius']            = radii
	# from tcrdist.public import TCRpublic
	# tp = TCRpublic(
	# 	tcrrep = tr, 
	# 	output_html_name = "quasi_public_clones.html")
	# tp.fixed_radius = False
	# rp = tp.report()






	

	# # Find 
















import pytest
import numpy as np
import pandas as pd
import os
import numpy as np
from scipy import sparse

from tcrdist.memory import _partition
from tcrdist.memory import collapse_csrs
from tcrdist.memory import gen_sparse_rw_on_fragment
from tcrdist.repertoire import TCRrep
from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory

def test_partion_size_2_on_6():
	"""
	UNIT TEST _partition
	"""
	assert _partition([1,2,3,4,5,6], 2) == [[1, 2], [3, 4], [5, 6]]

def test_partion_size_3_on_6():
	assert _partition([1,2,3,4,5,6], 3) == 	[[1, 2, 3], [4, 5, 6]]

def test_partion_size_3_on_5():
	assert _partition([1,2,3,4,5], 3) == 	[[1, 2, 3], [4, 5]]


def test_collapse_csrs_axis_0():
	"""
	UNIT TEST _collapse_csrs (axis = 0)
	"""
	A = np.array([[1, 0, 0, 1, 0, 0], [0, 0, 2, 0, 0, 1], [0, 0, 0, 2, 0, 0]])
	B = np.array([[3, 3, 3, 1, 0, 0], [5, 5, 2, 0, 0, 1], [6, 7, 0, 2, 0, 0]])
	A_s = sparse.csr_matrix(A)
	B_s = sparse.csr_matrix(B)
	sparse.save_npz("A.npz", A_s)
	sparse.save_npz("B.npz", B_s)
	AB_s = collapse_csrs(["A.npz","B.npz"], axis = 0)
	AB = AB_s.todense()
	assert np.all(AB == np.concatenate([A,B], axis = 0))


def test_collapse_csrs_axis_1():
	"""
	UNIT TEST _collapse_csrs (axis = 1)
	"""
	A = np.array([[1, 0, 0, 1, 0, 0], [0, 0, 2, 0, 0, 1], [0, 0, 0, 2, 0, 0]])
	B = np.array([[3, 3, 3, 1, 0, 0], [5, 5, 2, 0, 0, 1], [6, 7, 0, 2, 0, 0]])
	A_s = sparse.csr_matrix(A)
	B_s = sparse.csr_matrix(B)
	sparse.save_npz("A.npz", A_s)
	sparse.save_npz("B.npz", B_s)
	AB_s = collapse_csrs(["A.npz","B.npz"], axis = 1)
	AB = AB_s.todense()
	assert np.all(AB == np.concatenate([A,B], axis = 1))

def test_gen_sparse_rw_on_fragment():
	"""
	UNIT TEST ON _gen_sparse_rw_on_fragment, 
		TESTS THAT A SPARSE REPRESENTATION MATRIX IS COMPUTED, SAVED TO .npz 
		AND RETURNED AS CORRECT DIMENSIONS
	"""
	df = pd.read_csv("dash.csv")
	tr = TCRrep(cell_df = df,               #(2)
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv',
	            compute_distances = False)
	r = gen_sparse_rw_on_fragment(tcrrep = tr, ind = list(range(0,10)), outfile = "TESTONLY.csr", matrix_name = "rw_beta", max_distance = 1000)
	assert r == True
	assert os.path.isfile("TESTONLY.csr.npz")
	sparse_matrix_fragment = sparse.load_npz("TESTONLY.csr.npz")
	assert sparse_matrix_fragment.todense().shape[0] == 10
	assert sparse_matrix_fragment.todense().shape[1] == tr.clone_df.shape[0]
















def test_TCRpublic_with_neighborhood_dif():
    """
    Use values from neighborhood_diff
    """
    import os
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import TCRpublic  
    fn = os.path.join('tcrdist','data','covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.radius.csv')
    df = pd.read_csv(fn)
    tr = TCRrep(cell_df = df[['cohort','subject','v_b_gene', 'j_b_gene','cdr3_b_aa', 'radius']], 
                organism = "human", 
                chains = ["beta"])
    
    from tcrdist.rep_diff import neighborhood_diff
    ndif = neighborhood_diff(   clone_df= tr.clone_df, 
                                    pwmat = tr.pw_beta, 
                                    count_col = 'count', 
                                    x_cols = ['cohort'], 
                                    knn_radius = 25, 
                                    test_method = "chi2")
    # Add neighbors and other columns of interest 
    # from neighbor_diff result to the clone_df
    tr.clone_df = pd.concat([tr.clone_df, ndif[['neighbors', 'K_neighbors','val_0','ct_0','pvalue']] ], axis = 1)
    # Because neighors and K_neighbor are already added to the clone_df 
    # TCRpublic.report() uses those instead of finding new ones.
    tp = TCRpublic(
        tcrrep = tr, 
        output_html_name = "quasi_public_clones_with_ndif.html")
    # Add any columns neighbor_diff columns 
    #that you want to display in the final report
    tp.labels.append('val_0')
    tp.labels.append('ct_0')
    tp.labels.append('pvalue')
    # chagne sort to be pvalue not publicity
    tp.sort_columns = ['pvalue']
    # because you are sorting by pvalue, change to True
    tp.sort_ascending = True
    tp.report()




from .gene_pairing_plot import plot_pairings
import os

__all__ = ['plot_pairings']

def _write_svg(svg, name, dest = '.'):
    fn = os.path.join(dest, name)
    with open(fn, 'w') as fh:
        fh.write(f"{svg}")

def _write_svgs(svgs, name, dest = '.'):
    """ 
    writes a svg string from multiple svgs to a file dest/name 
    
    Parameters 
    ----------
    svgs : list of strs
        list of svgs strings
    name : str
        filename to output svg composite string
    dest : str
        destination folder
    """
    fn = os.path.join(dest, name)
    with open(fn, 'w') as fh:
        fh.write(f"<body>\n")
        
        for svg in svgs:
            fh.write(f"<div>\n")
            fh.write(f"\t{svg}\n")
            fh.write(f"</div>\n")
        fh.write(f"\n</body>")   



# tabulation 
import os
import pandas as pd
import numpy as np
import re
import scipy.sparse

def tabulate(clone_df1, clone_df2, pwmat = None, cdr3_name = 'cdr3_b_aa', v_gene_name = 'v_b_gene', j_gene_name = 'j_b_gene'):
	"""
	Tabulates frequency and counts of biochemically 
	similar clones in a bulk clone DataFrame <clone_df2>
	to special cloens o interest in <clone_df1>, most likely 
	derived from antigen-enriched repertoires (MIRA or tetramer-based enrichments). 

	It is REQUIRED that clone_df1 contain a 'radius' and 'regex' column, 
	specifiying the maximum search radius per clone and motif constraint.

	Results are provided for 
		(1)             TCRDIST <=  radius 
		(2) (regex_adj) TCRDIST <= radius + CDR3 matches regex
		(3) (tcrdist0)  TCRDIST == 0, (in almost all cases a direct match of both TRV Gene and CDR3 AA)

	Parameters
	----------
	clone_df1 : pd.DataFrame
		contains the metaclonotype definitions (at a minimum: cdr3,cdr2,cdr1,phmc,vgene,radius,regex)
	clone_df2 : pd.DataFrame
		contains the bulk clones (at a minimum: cdr3,cdr2,cdr1,phmc,vgene)
	pwmat : np.ndarray or scipy.sparse.csr.csr_matrix
		
	max_radius : int 
		This refers to the maximum radius used overall in the sparse compression. 
		row-wise all zeros will be converted to this value + 1.
	cdr3_name : str
		must describe CDR3 in both clone_df1 and clone_df2 e.g., 'cdr3_b_aa'
	v_gene_name : str
		must describe TRV gene =in both clone_df1 and clone_df2e.g., 'v_b_gene'
	j_gene_name : str
		must describe TRJ gene in both clone_df1 and clone_df2 e.g., 'j_b_gene'
	
	Returns 
	-------
	result_df : pd.DataFrame

	"""
	assert 'radius' in clone_df1.columns, '<radius> must be a column of clone_df1'
	assert 'regex' in clone_df1.columns, '<radius> must be a column of clone_df1'
	assert isinstance(clone_df1, pd.DataFrame), 'clone_df1 must be a Pandas DataFrame'
	assert isinstance(clone_df2, pd.DataFrame), 'clone_df1 must be a Pandas DataFrame'
	assert isinstance(pwmat,np.ndarray) or isinstance(pwmat, scipy.sparse.csr.csr_matrix)
	assert pwmat.shape[0] == clone_df1.shape[0]
	assert pwmat.shape[1] == clone_df2.shape[0]
	print(f"MATRIX SHAPE {pwmat.shape}")
	print(f"MATRIX TYPE  {type(pwmat)}")
	
	max_radii = clone_df1['radius'].values
	max_radius = clone_df1['radius'].values.max() + 1

	# Each search sequence has a max radius 
	icol = list()
	icol_0 = list()
	idist = list()
	for i,radius in zip(range(pwmat.shape[0]), max_radii):
		# convert a row of sparse matrix to an array, replace zeros with max_radius, replace -1 with 0
		if isinstance(pwmat, scipy.sparse.csr.csr_matrix):
			row_distances = pwmat[i,].toarray()[0] #)[0]
			row_distances[row_distances == 0] = max_radius
			row_distances[row_distances == -1] = 0
		elif isinstance(pwmat, np.ndarray):
			row_distances = pwmat[i,] #)[0]
		else:
			raise TypeError('<pwma> must be a numpy.ndarray or scipy.sparse.csr.csr_matrix')

		columns_within_radius = list(np.where(row_distances <= radius)[0])
		columns_within_radius_0 = list(np.where(row_distances == 0)[0])
		distances_within_radius = list(row_distances[columns_within_radius])
		icol.append(columns_within_radius)
		idist.append(distances_within_radius)
		icol_0.append(columns_within_radius_0)
	
	# Retrieve sequences from the bulk clone_df
	iseqs   = [clone_df2[cdr3_name].iloc[x].to_list() for x in icol]
	ivgenes = [clone_df2[v_gene_name].iloc[x].to_list()  for x in icol]
	ijgenes = [clone_df2[j_gene_name].iloc[x].to_list()  for x in icol]
	# Retrieve abundances from the bulk clone df
	icounts    = [clone_df2['count'].iloc[x].to_list()                 for x in icol]
	ifreqs     = [clone_df2['productive_frequency'].iloc[x].to_list()  for x in icol]
	
	isumcounts    = [np.sum(x) for x in icounts]
	isumfreqs     = [np.sum(x) for x in ifreqs ]

	df_summ = pd.DataFrame({ 'bulk_sum_freq'     : isumfreqs,
							 'bulk_sum_counts'   : isumcounts,
							 'bulk_seqs'      	 : iseqs,
							 'bulk_v_genes'   	 : ivgenes,
							 'bulk_j_genes'   	 : ijgenes,
							 'bulk_distances' 	 : idist,
							 'bulk_counts'    	 : icounts, 
							 'bulk_freqs'     	 : ifreqs})
	result_df = pd.concat([clone_df1.copy(), df_summ], axis = 1)

	regex_on_bulk = [[re.search(pattern = r['regex'], string =  s) for s in r['bulk_seqs']] for _,r in result_df.iterrows()]
	bulk_regex_match = [[True if (x is not None) else False for x in sublist] for sublist in regex_on_bulk]
	result_df['bulk_regex_match'] = bulk_regex_match
	result_df['bulk_sum_freqs_regex_adj'] = [pd.Series(r['bulk_freqs'], dtype = "float64")[pd.Series(r['bulk_regex_match'], dtype = 'bool')].sum() for i,r in result_df.iterrows()]
	result_df['bulk_sum_counts_regex_adj'] = [pd.Series(r['bulk_counts'], dtype = "int32")[pd.Series(r['bulk_regex_match'],  dtype = 'bool')].sum() for i,r in result_df.iterrows()]

	# TCRDIST == 0
	icounts_0 = [clone_df2['count'].iloc[x].to_list() for x in icol_0]
	ifreqs_0 = [clone_df2['productive_frequency'].iloc[x].to_list()  for x in icol_0]
	isumcounts_0 = [np.sum(x) for x in icounts_0]
	isumfreqs_0     = [np.sum(x) for x in ifreqs_0 ]
	result_df['bulk_sum_freqs_tcrdist0'] = isumfreqs_0
	result_df['bulk_sum_counts_tcrdist0'] = isumcounts_0

	return result_df







def test_example_tree_args():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash.csv").sample(100, random_state=1).reset_index(drop = True)

    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    tcrtree = TCRtree(tcrrep = tr, 
          html_name = 'dash.mouse.b.tree.html')

    tcrtree.default_hcluster_diff_kwargs['x_cols'] = ['epitope']

    tcrtree.default_member_summ_kwargs['addl_cols'] : ['subject', 'epitope']

    tcrtree.default_plot_hclust_props['alpha_col'] = 'pvalue'
    tcrtree.default_plot_hclust_props['alpha'] = 1.0

    tcrtree.build_tree()


import pytest
import Levenshtein
def my_own_metric(s1,s2):   
    return Levenshtein.distance(s1,s2)
def test_example_7():
    """
    If you don't want to use OOP, but you I still want a multi-CDR 
    tcrdistances on a single chain, using you own metric 

    def my_own_metric(s1,s2):   
        return Levenshtein.distance(s1,s2)    
    """
    import multiprocessing
    import pandas as pd
    from tcrdist.rep_funcs import _pws, _pw
    
    df = pd.read_csv("dash2.csv")
    
    metrics_b = {
        "cdr3_b_aa" : my_own_metric ,
        "pmhc_b_aa" : my_own_metric ,
        "cdr2_b_aa" : my_own_metric ,
        "cdr1_b_aa" : my_own_metric }

    weights_b = { 
        "cdr3_b_aa" : 1,
        "pmhc_b_aa" : 1,
        "cdr2_b_aa" : 1,
        "cdr1_b_aa" : 1}

    kargs_b = {  
        'cdr3_b_aa' : 
            {'use_numba': False},
        'pmhc_b_aa' : {
            'use_numba': False},
        'cdr2_b_aa' : {
            'use_numba': False},
        'cdr1_b_aa' : {
            'use_numba': False}
        }

    dmats =  _pws(df = df , 
                metrics = metrics_b, 
                weights = weights_b, 
                kargs   = kargs_b , 
                cpu     = 1, 
                uniquify= True, 
                store   = True)

    print(dmats.keys())


"""
Functions that automate expressive functions of tcrdist3. 

These use sensible defaults but users should consider 
flexibile elements of these functions that might better 
suit their specific data or research questions. 
"""
import os
import pandas as pd
import numpy as np
import warnings
from progress.bar import IncrementalBar

"""
Test Automatic Pgen Esimator Functions
"""
def auto_pgen(tcrrep):
	"""
	Automate a pgen estimation of cdr3s alpha/beta given a tcrrep with a clones_df attribute

	Parameters
	----------
	tcrrep : tcrdist.repertoire.TCRrep
		TCRrep instance with a clone_df

	Returns
	-------
	tcrrep : tcrdist.repertoire.TCRrep

	tcrrep : tcrdist.repertoire.TCRrep
		TCRrep instance with a clone_df
	"""
	if tcrrep.organism == "mouse" and 'alpha' in tcrrep.chains:
		raise ValueError("UNFORTUNATELY OLGA DOES NOT YET SUPPORT MOUSE ALPHA PGEN ESTIMATES")

	for chain in tcrrep.chains:
		tcrrep = _auto_pgen(tcrrep=tcrrep, organism = tcrrep.organism, chain = chain, ncpus = tcrrep.cpus)
	return tcrrep


def _auto_pgen(tcrrep = None, organism = 'human', chain = 'beta', ncpus = 2):
	"""
	Automate a pgen estimation of cdr3s alpha/beta given a tcrrep with a clones_df attribute

	Parameters
	----------
	tcrrep : tcrdist.repertoire.TCRrep
		TCRrep instance with a clone_df
	organism : str
		'human' or 'mouse'
	chain : str
		'beta' or 'alpha'

	Returns
	-------
	tcrrep : tcrdist.repertoire.TCRrep
	"""
	import tcrdist
	import parmap
	import pandas as pd
	from tcrdist.pgen import OlgaModel
	assert organism in ['human', 'mouse']
	assert chain in ['beta', 'alpha']
	assert isinstance(tcrrep, tcrdist.repertoire.TCRrep)
	assert isinstance(tcrrep.clone_df, pd.DataFrame)

	cdr3_col = {'alpha': 'cdr3_a_aa', 'beta': 'cdr3_b_aa'}[chain]
	cdr3s = tcrrep.clone_df[cdr3_col]

	olga_models = { 
		('human', 'beta') : 
			OlgaModel(chain_folder = "human_T_beta", recomb_type="VDJ"),
		('human', 'alpha') :
			OlgaModel(chain_folder = "human_T_alpha", recomb_type="VJ"),
		('mouse', 'beta') : 
			OlgaModel(chain_folder = "mouse_T_beta", recomb_type="VDJ")}
	
	olga_model = olga_models[(organism, chain)]

	pgens = parmap.map(olga_model.compute_aa_cdr3_pgen, cdr3s, pm_pbar = True, pm_processes = ncpus)
	tcrrep.clone_df[f"pgen_{cdr3_col}"] = pgens

	return tcrrep
	


def _default_sampler(organism = 'human', chain = 'beta'):
	assert organism in ['human', 'mouse']
	assert chain in ['beta','alpha']

	default_tcrsampler_generator = {
		('human','beta'): 
			_default_tcrsampler_human_beta,
		('human','alpha'): 
			_default_tcrsampler_human_alpha,
		('mouse','beta'): 
			_default_tcrsampler_mouse_beta,
		('mouse','alpha'): 
			_default_tcrsampler_mouse_alpha, 
		}[(organism, chain)]
	
	return default_tcrsampler_generator



def _default_tcrsampler_human_beta(default_background = None, default_background_if_missing=None):
	"""
	Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

	Returns
	-------
	t : tcrsampler.sampler.TCRsampler 
	"""
	from tcrsampler.sampler import TCRsampler
	if default_background is None:
		default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv'
		
	if default_background_if_missing is None:
		default_background_if_missing ='britanova_human_beta_t_cb.tsv.sampler.tsv.zip'
	
	
	print(default_background)

	try: 
		t = TCRsampler(default_background=default_background)
	except OSError:
		t = TCRsampler()
		t.download_background_file(default_background_if_missing)
		t = TCRsampler(default_background=default_background)
	return t

def _default_tcrsampler_human_alpha(default_background = None, default_background_if_missing=None ):
	"""
	Responsible for providing the default human alpha sampler 'ruggiero_human_alpha_t.tsv.sampler.tsv'
	"""
	from tcrsampler.sampler import TCRsampler
	if default_background is None:
		default_background = 'ruggiero_human_alpha_t.tsv.sampler.tsv'
	if default_background_if_missing is None:
		default_background_if_missing =  'ruggiero_human_alpha_t.tsv.sampler.tsv.zip'
	
	print(default_background)

	try: 
		t = TCRsampler(default_background=default_background)
	except OSError:
		t = TCRsampler()
		t.download_background_file(default_background_if_missing)
		t = TCRsampler(default_background=default_background)
	return t

def _default_tcrsampler_mouse_beta(default_background = None, default_background_if_missing=None):
	"""
	Responsible for providing the default mouse beta sampler

	Returns
	-------
	t : tcrsampler.sampler.TCRsampler 
	"""
	from tcrsampler.sampler import TCRsampler

	if default_background is None:
		default_background = 'ruggiero_mouse_beta_t.tsv.sampler.tsv'
	if default_background_if_missing is None:
		default_background_if_missing =  'ruggiero_mouse_sampler.zip'

	print(default_background)

	try: 
		t = TCRsampler(default_background=default_background)
	except OSError:
		t = TCRsampler()
		t.download_background_file(default_background_if_missing)
		t = TCRsampler(default_background=default_background)
	return t

def _default_tcrsampler_mouse_alpha(default_background = None, default_background_if_missing=None):
	"""
	Responsible for providing the default mouse alpha sampler
	"""
	from tcrsampler.sampler import TCRsampler
	
	if default_background is None:
		default_background = 'ruggiero_mouse_alpha_t.tsv.sampler.tsv'
	if default_background_if_missing is None:
		default_background_if_missing =  'ruggiero_mouse_sampler.zip'
	
	print(default_background)

	try: 
		t = TCRsampler(default_background=default_background)
	except OSError:
		t = TCRsampler()
		t.download_background_file(default_background_if_missing)
		t = TCRsampler(default_background=default_background)
	return t



import pytest
import pandas as pd
import numpy as np
import os

def test__default_tcrsampler_human_beta():
    from tcrdist.automate import _default_tcrsampler_human_beta
    from tcrsampler.sampler import TCRsampler
    r = _default_tcrsampler_human_beta()
    assert isinstance(r, TCRsampler)

def test__default_tcrsampler_human_alpha():
    from tcrdist.automate import _default_tcrsampler_human_alpha
    from tcrsampler.sampler import TCRsampler
    r = _default_tcrsampler_human_alpha()
    assert isinstance(r, TCRsampler)

def test__default_tcrsampler_mouse_beta():
    from tcrdist.automate import _default_tcrsampler_mouse_beta
    from tcrsampler.sampler import TCRsampler
    r = _default_tcrsampler_mouse_beta()
    assert isinstance(r, TCRsampler)

def test__default_tcrsampler_mouse_alpha():
    from tcrdist.automate import _default_tcrsampler_mouse_alpha
    from tcrsampler.sampler import TCRsampler
    r = _default_tcrsampler_mouse_alpha()
    assert isinstance(r, TCRsampler)

"""
Test Automatic Pgen Esimator Functions
"""
def test__auto_pgen_mouse_beta():
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv").sample(10, random_state = 3)
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    from tcrdist.automate import _auto_pgen
    tr = _auto_pgen(tr, chain = 'beta', organism = 'mouse')
    assert isinstance(tr.clone_df.pgen_cdr3_b_aa, pd.Series)

    
def test_auto_pgen_mouse_beta():
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv").sample(10, random_state = 3)
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    from tcrdist.automate import auto_pgen
    tr = auto_pgen(tr)
    assert isinstance(tr.clone_df.pgen_cdr3_b_aa, pd.Series)

def test_auto_pgen_mouse_alpha_beta_ValueError():
    """ If auto_pgen called on TCRrep with organism == mouse and 
    chain including 'alpha', Raises ValueError"""

    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv").sample(10, random_state = 3)
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    from tcrdist.automate import auto_pgen
    with pytest.raises(ValueError):
        tr = auto_pgen(tr)
        
def test__auto_pgen_human_alpha_beta():
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    df = pd.read_csv("dash_human.csv").sample(10, random_state = 3)
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
   
    from tcrdist.automate import _auto_pgen
    tr = _auto_pgen(tr, chain = 'beta', organism = 'human')
    assert isinstance(tr.clone_df.pgen_cdr3_b_aa, pd.Series)
    assert isinstance(tr.clone_df.pgen_cdr3_b_aa[0], float)
    tr = _auto_pgen(tr, chain = 'alpha', organism = 'human')
    assert isinstance(tr.clone_df.pgen_cdr3_a_aa, pd.Series)
    assert isinstance(tr.clone_df.pgen_cdr3_a_aa[0], float)

def test_auto_pgen_human_alpha_beta():
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    df = pd.read_csv("dash_human.csv").sample(10, random_state = 3)
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    from tcrdist.automate import auto_pgen
    tr = auto_pgen(tr)
    assert isinstance(tr.clone_df.pgen_cdr3_b_aa, pd.Series)






import pytest 
from tcrdist import setup_tests #get_url, download_and_extract_zip_file
from tcrdist import paths
import os

def test_get_url():
    url = setup_tests.get_url("bulk.csv.zip", source = 'dropbox')
    assert url == "https://www.dropbox.com/s/g6k2h1ed5d5sabz/bulk.csv.zip?dl=1"

def test_download_and_extract_zip_file():
    setup_tests.download_and_extract_zip_file(zipfile = "sant.csv.zip", source = 'dropbox', dest = paths.path_to_base)
    # assert download succeed
    assert os.path.isfile(os.path.join(paths.path_to_base, "sant.csv.zip"))
    # assert download + unzip succeed
    assert os.path.isfile(os.path.join(paths.path_to_base, "sant.csv"))
    



import numpy as np
import pandas as pd
from scipy import stats

__all__ = ['generalized_simpsons_entropy',
           'simpsons_difference',
           'effective_number',
           'fuzzy_diversity']

def _calc_Z_base(counts, r_vec):
    """Python function for computing Z that is compatible with
    numba compilation."""
    n = np.sum(counts)
    
    K = len(counts)
    p = counts / n
    
    Z_out = np.zeros(len(r_vec))
    for r_i in prange(len(r_vec)):
        r = r_vec[r_i]
        Z = 0
        for i in range(K):
            if counts[i] == 0:
                continue
            else:
                prod = 1
                for k in range(1, r):
                    prod *= 1 - (counts[i] - 1) / (n - k)
                Z += prod * p[i]
        Z_out[r_i] = Z
    return Z_out

def _calc_stdev_base(counts, r_vec):
    """Python function for computing Z standard deviation that is
    compatible with numba compilation."""
    n = np.sum(counts) # samples1
    p = counts / n
    K = len(p)

    stdev = np.zeros(len(r_vec))
    for r_i in prange(len(r_vec)):
        r = r_vec[r_i]
        """Assumes that the last count is not 0"""
        # index = c1.shape[0] - 1
        h_hat = np.zeros(K - 1)
        tmp = (1 - p[:-1])**r + r*p[:-1]*(1 - p[:-1])**(r-1) - (1 - p[-1])**r - r*p[-1]*(1 - p[-1])**(r-1)
        h_hat[p[:-1] > 0] = tmp[p[:-1] > 0]

        sigma = np.zeros((K-1, K-1))
        for i in range(K-1):
            for j in range(K-1):
                if i == j:
                    sigma[i, j] = p[i] * (1 - p[i])
                else:
                    sigma[i, j] = -p[i] * p[j]
        
        v = np.dot(np.dot(h_hat.T, sigma), h_hat)
        """
        v = 0
        for i in range(K-1):
            for j in range(K-1):
                v += h_hat[i] * sigma[i, j] * h_hat[j]
        """
        stdev[r_i] = np.sqrt(v)
    return stdev

try:
    """Try to import and compile using jit, otherwise fall back on
    numpy and python loops (slow for large datasets)"""
    from numba import jit, prange

    _calc_Z = jit(_calc_Z_base, nopython=True, error_model='numpy', parallel=True)
    _calc_stdev = jit(_calc_stdev_base, nopython=True, error_model='numpy', parallel=True)
except ImportError:
    prange = range
    _calc_Z = _calc_Z_base
    _calc_stdev = _calc_stdev_base

def simpsons_difference(counts1, counts2, orders=[2], alpha=0.05):
    """Difference in diversity between two communities.

    Parameters
    ----------
    counts1, counts2 :  : np.ndarray or pd.Series
        Vector of counts for each species.
    orders : np.ndarray or list of integers
        Order for calculation. r = 2 is equivalent to common Simpson's entropy computations.
        Increasing r gives more relative importance to rare species.
    alpha : float
        Upper and lower confidence levels define the 1 - alpha/2 confidence interval

    Returns
    -------
    difference, lcl, ucl : np.ndarray, shape len(orders)"""
    orders = np.asarray(orders).astype(float)

    counts1 = np.asarray(counts1).astype(float)
    n1 = np.sum(counts1)

    counts2 = np.asarray(counts2).astype(float)
    n2 = np.sum(counts2)

    Z1 = _calc_Z(counts1, orders)
    sdev1 = _calc_stdev(counts1, orders)

    Z2 = _calc_Z(counts2, orders)
    sdev2 = _calc_stdev(counts2, orders)

    d = Z1 - Z2

    criticalz = -stats.norm.ppf(alpha / 2)
    
    lcl = d - criticalz * np.sqrt(sdev1**2 / n1 + sdev2**2 / n2)
    ucl = d + criticalz * np.sqrt(sdev1**2 / n1 + sdev2**2 / n2)
    return d, lcl, ucl

def effective_number(Z, orders=[2]):
    """Effective number of species is the number of equiprobable
    species that would yield the same diversity as a given distribution.

    As it is a monotonic transformation this function can be used to
    transform confidence intervals on Z as well.

    Parameters
    ----------
    Z : np.ndarray of floats, [0, 1]
        Generalized Simpson's entropies
    orders : np.ndarray of integers
        Order of the generalized Simpson's entropy.
        Must match the orders used in the calculation of Z's.

    Returns
    -------
    D : float
        Effective number"""
    orders = np.asarray(orders).astype(float)
    return 1 / (1 - Z**(1 / orders))

def generalized_simpsons_entropy(counts, orders=[2], alpha=0.05):
    """Generalized Simpson’s entropy of order r can be interpreted as
    the average information brought by the observation of an individual/species.
    
    Its information function I(p) = (1 − p)*r represents the probability of
    not observing a single individual of a species with proportion p in a sample
    of size r. Thus I is an intuitive measure of rarity.

    Above is quoted from:
    Grabchak M, Marcon E, Lang G, Zhang Z (2017) The generalized Simpson’s entropy
        is a measure of biodiversity. PLoS ONE 12(3): e0173305.
        https://doi.org/10.1371/journal.pone.0173305

    It is common to evaluate using a series of orders as long as r < len(counts) - 1

    Parameters
    ----------
    counts : np.ndarray or pd.Series
        Vector of counts for each species.
    orders : np.ndarray of integers
        Order for calculation. r = 2 is equivalent to common Simpson's entropy computations.
        Increasing r gives more relative importance to rare species.
    alpha : float
        Upper and lower confidence levels define the 1 - alpha/2 confidence interval

    Returns
    -------
    Z, LCL, UCL : floats or as pd.Series if counts is a pd.Series
        Generalized Simpson's entropy Z and lower (upper) confidence limits."""
    orders = np.asarray(orders).astype(float)

    if type(counts) is pd.Series:
        return_series = True
        name = counts.name
        counts = counts.values.astype(float)
    else:
        counts = np.asarray(counts).astype(float)
    n = np.sum(counts)

    Z = _calc_Z(counts, orders)
    sdev = _calc_stdev(counts, orders)
    criticalz = -stats.norm.ppf(alpha / 2)

    lcl = Z - criticalz * sdev / np.sqrt(n)
    ucl = Z + criticalz * sdev / np.sqrt(n)

    if return_series:
        return pd.DataFrame({'order':orders.astype(int), 'Z':Z, 'Z_LCL':lcl, 'Z_UCL':ucl,
                              'D':effective_number(Z, orders),
                              'D_LCL':effective_number(lcl, orders),
                              'D_UCL':effective_number(ucl, orders)}).set_index('order')
    else:
        return Z, lcl, ucl

def _non_generalized_simpsons_index(counts):
    p = counts / np.sum(counts)
    D = (p * p).sum()
    return 1 - D


def fuzzy_diversity(counts, pwmat, order=2, threshold=1, nsamples=1000, force_sampling=False):
    """Compute a "fuzzy" diversity index that takes into account the number of similar
    members in a community, not just the ones that are identically matched.

    Idea published as "TCRdiv" in Dash et al. (2017) in the area of T cell receptor
    (TCR) analysis. The community is a TCR repertoire and the members are unique
    TCR clones.

    The diversity index is similar to Simpson's diversity index (SDI) in that it estimates
    the chance of sampling m members from the community that are all more similar
    to each other than some pre-specified threshold.

    Fuzzy diversity is the sum of the probability of n members that match exactly,
    plus the probability of m members matching within the threshold. Or it can be expressed
    as 1 - probability of drawing m members that are not similar (distance > threshold).

    For a confidence interval use some kind of bootstrap approach.

    Parameters
    ----------
    counts : np.ndarray [n]
        Number of observations of each member in the community
    pwmat : np.ndarray [n x n]
        Pairwise distances (not similarities) for all members in the community.
        Consider using 1 to represent members that are not similar, and 0 for others,
        or use the threshold parameter.
    order : int
        Order of the SDI, i.e. the number of members sampled to see if they are
        similar to one another
    threshold : float
        A distance threshold that defines whether two members are similar to one
        another: similar if distance < threshold
    force_sampling : bool
        Force use of the sampling approach, even for order=2
        Only useful for checking the order=2 result.

    Returns
    -------
    div : float
        Estimate of fuzzy diversity"""

    counts = np.asarray(counts)[:, None]

    n = np.sum(counts)
    
    cts = counts * counts.T

    assert cts.shape[0] == pwmat.shape[0]
    assert cts.shape[0] == pwmat.shape[1]

    up = np.triu_indices_from(cts, k=1)

    if order == 2 and not force_sampling:
        """The pwmat represents drawing two members and seeing if they are similar
        so order = 2 is just a simple summary of the pairwise distance matrix:
        the proportion of distance pairs that represent similar members"""
        div = np.sum(cts[up] * (pwmat <= threshold)[up]) / np.sum(cts)
    else:
        """Higher orders could be computed using a sampling approach
        since its not scalable to compute all possible triplets, etc.
        to see which are all similar. A triplet consists of 3 pairwise distances
        that must all be within the threshold to be a similar triplet."""
        dvec = (pwmat <= threshold)[up]
        prob = cts[up] / np.sum(cts[up])
        samples = np.random.choice(dvec, size=(order, nsamples), replace=True, p=prob)
        
        div = np.mean(np.all(samples, axis=0))

    return 1 - div








"""
centers 

Module contains functions for evaluating TCRs as center(oids) of meta-clonotypes.

find_center 
"""
import warnings 
import numpy as np
from tcrdist.ecdf import distance_ecdf

def calc_radii(tr, tr_bkgd, chain = 'beta', ctrl_bkgd = 10**-5, use_sparse = True, max_radius=50, chunk_size=100, **kwargs):
	"""
	Simply find maximum radii based on an antigen enriched repertoires <tr> 
	and a background antigen-naive background. 

	IMPORTANT: This function will work with a precomputed TCRrep.rw_chain matrix if it 
	is supplied and matches row dimensions of tr_bkgd. This way the user does 
	not have to recompute it each time they with to change ctrl_bkgd. It 
	also allows them to customize the distance prior to this stage, although 
	**kwargs can be passed to the compute_sparse_rect_distances or 
	compute_rect_distances function

	Parameters
	----------
	tr : tcrdist.repertoire.TCRrep
		An antigen enriched repertoires TCRrep object
	tr_bkgd : tcrdist.repertoire.TCRrep
		A background antigen-naive background TCRrep object
	chain : str
		e.g, 'beta', ctrl_bkgd = 10**-5, use_sparse = True, max_radius=50, chunk_size=50
	ctrl_bkgd : float
		e.g., 10**-5, 
	use_sparse : bool 
		If True, uses a sparse implementation, 
	max_radius : int 
		Values beyond this max_radius are set to 0 in sparcification
	chunk_size : int
		How many rows to process at a time, based on memory available (100 is a good default, but for more see notes)
	**kwargs will be passed to either compute_sparse_rect_distances or compute_rect_distances
	Returns
	-------
	max_radii : list
		List of length equal to the numer of rows in <tr> clond_df. 
		These are radius at which the number of expected background 
		sequences are 'controlled' at a rate of <ctlr_bkgd> 

	Notes
	-----
	TODO: Discuss chunk size and memory
	"""
	assert chain in ['alpha','beta','gamma', 'delta']
	if 'weights' not in tr_bkgd.clone_df.columns:
		warnings.warn("No weights provided in background repertoire, setting to 1")
		tr_bkgd.clone_df['weights'] = 1

	# USER MAY HAVE ALRRADY COMPUTED TCRrep.rw_
	if getattr(tr, f"rw_{chain}", None) is not None:
		if getattr(tr, f"rw_{chain}").shape[1] == tr_bkgd.clone_df.shape[0]:
			print(f"IT APPEARS THAT (TCRrep.rw_{chain}) HAS ALREDY BEEN COMPUTED AND MATCHES BACKGROUND TCRrep SIZE")
			print(f"USING EXISTING (TCRrep.rw_{chain}). SET TCRrep.rw_{chain} = None IF YOU WANT TO RECOMPUTE IT.")
	else:
		if use_sparse:
			print(f"COMPUTING SPARSE RECT MATRIX TO FIND RADIUS: (TCRrep.rw_{chain})")
			print(f"USING {tr.cpus} CPUS")
			tr.compute_sparse_rect_distances(df = tr.clone_df, 
											 df2 = tr_bkgd.clone_df,
											 radius=max_radius,
											 chunk_size=chunk_size,
											 **kwargs)
		else:
			print(f"COMPUTING FULL RECT MATRIX TO FIND RADIUS, (TCRrep.rw_{chain})")
			tr.compute_rect_distances(df = tr.clone_df, 
									  df2 = tr_bkgd.clone_df, 
									  store = False,
									  **kwargs)
	print(f"COMPUTING ECDFS PER TCR, TO FIND APPROPRIATE MAX RADII AT {ctrl_bkgd}")
	thresholds, ecdfs = distance_ecdf(pwrect = getattr(tr, f"rw_{chain}"), 
						  thresholds = np.array(range(0,max_radius, 2)), 
						  weights= tr_bkgd.clone_df.weights, 
						  pseudo_count=0, 
						  skip_diag = False, 
						  absolute_weight = True)
	# Based on acceptable ctrl_bkgd, we find max acceptable radi from each TCR
	#import pdb; import pdb; pdb.set_trace()
	all_radii = [pd.Series(x, index = thresholds) for x in ecdfs]
	max_radii = [s[s<=ctrl_bkgd].last_valid_index() for s in all_radii]
		# WARNING: There is a potential BUG in the output of the above line. 
		# That is, iF a radius is None (the next line will fail, thus set Nones to 0.
	max_radii = [i if (i is not None) else 0 for i in max_radii]
	print(f"RETURNING LIST OF MAX RADII")
	
	return max_radii, thresholds, ecdfs 







def find_centers_beta(	
	background_filename,
	target_filename,
	ncpus,
	min_nsubject,
	ctrl_bkgd = 10**-5, 
	prefilter = False):
	import os
	import pandas as pd
	import numpy as np 
	from tcrdist.repertoire import TCRrep
	from tcrdist.neighbors import compute_ecdf, bkgd_cntl_nn2
	from tcrdist.automate import auto_pgen
	from tcrdist.rep_diff import neighborhood_diff
	from tcrdist.summarize import _summ, _dist_summ, _select, filter_gt, filter_is, test_for_subsets, test_for_almost_subsets
	import scipy.sparse
	
	df_background       =   pd.read_csv(background_filename)
	print(df_background)
	tr_background       =   TCRrep( cell_df = df_background.copy(), 
									organism = "human", 
									chains= ['beta'], 
									compute_distances = False)

	df_mira             =   pd.read_csv(target_filename)
	df_mira             =   df_mira[['subject','cell_type','v_b_gene', 'j_b_gene', 'cdr3_b_aa']]
	print(df_mira)
	tr                  =   TCRrep(	cell_df = df_mira.copy(), 
									organism = 'human', 
									chains = ['beta'], 
									db_file = 'alphabeta_gammadelta_db.tsv',
									store_all_cdr = False,
									compute_distances = True)
	
	if prefilter:
		# We can greatly cut down on the number of searches if we drop centroids without minimum publicicity
		nn_df = neighborhood_diff(clone_df= tr.clone_df,
								  pwmat = tr.pw_beta,
								  count_col = 'count',
								  x_cols = ['cell_type'],
								  knn_radius = 37)
		def tabulate_publicity(neighbor_df, clone_df, col_nn ='neighbors'):
			# Tabulate the number of unique subjects at each node
			neighbor_df['nsubject'] = neighbor_df[col_nn].apply( lambda x: len(set(_select(clone_df, iloc_rows =x, col = 'subject'))))
			return neighbor_df
		print(f"TABULATING PUBLIC CLUSTERS")
		nn_df = tabulate_publicity(nn_df, tr.clone_df)
		nn_df = filter_gt(nn_df, 'nsubject' , min_nsubject)
	
		if nn_df.shape[0] == 0:
			centers_df = pd.DataFrame({}, columns =  ['cdr3_b_aa','v_b_gene','j_b_gene','pgen','max_radi','target_hits','bkgd_hits','bkgd_hits_weighted','bkgd_total','ctrl','ctrl_weighted','target_misses','TR','TR2','BR_weighted','RR_weighted','OR_weighted','chi2dist','target_neighbors','target_seqs','background_neighbors','background_seqs','background_v','background_j','regex','target_re_hits','bkgd_re_hits','bkgd_re_weighted_hits','TR_re','BR_re_weighted','RR_re_weighted','OR_re_weighted','chi2re','chi2joint','nsubject'])
			tr.pw_beta[tr.pw_beta == 0]  = 1 # set true zeros to 1
			tr.pw_beta[tr.pw_beta > 50]  = 0 # ignores everything less than 100
			pw_beta_sparse = scipy.sparse.csr_matrix(tr.pw_beta)	
			return centers_df, pw_beta_sparse
	
		tr.clone_df = tr.clone_df.loc[nn_df.index, :].reset_index(drop = True)
		del nn_df
		# Compute pairwise again with filtered set
		tr.compute_distances()
		# compute pgens automatically, currently parmap will max out cpus on this step 
		
		

	print("COMPUTING PROBABILITY OF GENERATION")
	auto_pgen(tr)
	print(f"COMPUTING RECT DIST {tr.clone_df.shape[0]}x{tr_background.clone_df.shape[0]}")
	tr.compute_rect_distances(df = tr.clone_df, 
							  df2 = tr_background.clone_df, 
							  store = False)

	assert tr.rw_beta.shape[0] == tr.clone_df.shape[0]

	centers_df = bkgd_cntl_nn2(	tr = tr, 
								tr_background = tr_background,
								ctrl_bkgd = ctrl_bkgd,  #ctrl_bkgd = 2*10**-5
								weights =tr_background.clone_df.weights,
								col = 'cdr3_b_aa',
								ncpus = ncpus,
								thresholds = [x for x in range(0,38,2)], # Settign 38 as the max radius
								generate_regex = True, 
								test_regex = True)

	def tabulate_publicity(neighbor_df, clone_df, col_nn ='neighbors'):
		# Tabulate the number of unique subjects at each node
		neighbor_df['nsubject'] = neighbor_df[col_nn].apply( lambda x: len(set(_select(clone_df, iloc_rows =x, col = 'subject'))))
		return neighbor_df

	centers_df = tabulate_publicity(neighbor_df = centers_df, clone_df = tr.clone_df, col_nn ='target_neighbors')
		
	tr.rw_beta[tr.rw_beta == 0]  = 1 # set true zeros to 1
	tr.rw_beta[tr.rw_beta > 50]  = 0 # ignores everything less than 100
	rw_beta_sparse = scipy.sparse.csr_matrix(tr.rw_beta)
	#scipy.sparse.save_npzz(output_matrix_filename, rw_beta_sparse)
	return centers_df, rw_beta_sparse


def rank_centers(centers_filename = None, centers_df = None, rank_column = 'chi2joint', min_nsubject = 2, min_nr = 1):
	"""
	This function takes the output of tcrdist.neighbors.bkgd_cntl_nn2(), 
	a set of scored metaclonotypes (centers - TCRs + radius) 
	and ranks them by chi2 statistics, 
	prioritizing those that include lots of target sequences 
	while minimizing inclusion of background sequences. 
	
	Parameters
	----------
	centers_filename : str or None
		User can only provide centers_df or centers_filename but not both
		The filepath to a file containing metaclonotype centers information, generally produced with 
		tcrdist.neighbors.bkgd_cntl_nn2()
	centers_df : DataFrame or None
		User can only provide centers_df or centers_filename but not both.
		The Pandas DataFraem containing metaclonotype centers information, generally produced with 
		tcrdist.neighbors.bkgd_cntl_nn2()
	rank_column : str
		Default : 'chi2joint' (or 'chi2joint' (radius+motif averaged) or chi2re'(using motif only), 'chi2dist' (using radius only) 
	min_nsubject : int
		Default 2, (minimum publicity of the meta-clonotype). 
		That is, the minimum number of unique subjects contributing TCRs 
		among a group of biochemically TCRs to form a meta-clonotype. 
	min_nr : int
		Default 1, (minimum non-redundancy). Once the metaclonotypes are ranked, 
		the function requires that lower ranked meta-clonotypes to have a minimum number
		<min_nr> of new sequences not already spanned by a higher ranked meta-clonotype. 
	
	Returns
	-------
	df : DataFrame
	
	"""
	import pandas as pd
	import ast
	from tcrdist.summarize import filter_gt, filter_is, test_for_subsets, test_for_almost_subsets
	
	if centers_filename is not None and centers_df is not None:
		raise ValueError("rank centers can use <centers_filename> or <centers_df> but not both")
	if centers_df is None:
		df = pd.read_csv(centers_filename)
	else: 
		df = centers_df.copy()

	# VERY IMPORTANT NOTE, pandas reads lists as strings '[1,2]'; so we use ast.literal_eval to convert back t a list 
	if not isinstance(df['target_neighbors'][0], list):
		df['target_neighbors'] = df['target_neighbors'].apply(lambda s: list(ast.literal_eval(s)))
	df = df.sort_values(rank_column, ascending = False)
	df['novel'] = test_for_almost_subsets(df['target_neighbors'], min_nr)
	df = filter_gt(df, 'nsubject', min_nsubject).copy()
	df = filter_is(df, 'novel', 1).copy()
	return df


import re
import os
import pandas as pd
import numpy as np
from tcrdist.repertoire import TCRrep
from tcrdist.adpt_funcs import _valid_cdr3
import scipy.sparse
import ast

def check_tsv_csv(filename, check_column_names = ['cdr3_b_aa', 'v_b_gene', 'j_b_gene']):
	"""
	To avoid problems with .tsv or .csv, check for appropriate seperator based on expected columns
	"""
	#shape_csv = pd.read_csv(filename, sep = ",").shape
	columns_csv = pd.read_csv(filename, sep = ",").columns
	#shape_tsv = pd.read_csv(filename, sep = "\t").shape
	columns_tsv = pd.read_csv(filename, sep = "\t").columns
	
	if set(check_column_names) - set(columns_csv) == set():
		sep = ","
	elif set(check_column_names) - set(columns_tsv) == set():
		sep = "\t"
	else: 
		sep = False
		raise IOError("File provided does not appear to be either a .csv or .tsv or lacks required columns {check_column_names}")
	return sep


def centers_v_bulk(search_filename, bulk_filename, sep_search_filename = "\t"):
	"""
	"""
	# Get appropriate seperator (in this case we expect "\t", trust than verify)
	sep_bulk_filename = check_tsv_csv(filename = bulk_filename, check_column_names = ['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'templates', 'productive_frequency','count'])
	bulk_df = pd.read_csv(bulk_filename, sep = sep_bulk_filename)

	# Adaptive uses the terms templates, which are synonymous with counts in our nomenclature
	bulk_df['count'] = bulk_df['templates'].copy()
	# Ensure that all bulk sequences have valid cdr3s
	v = bulk_df.cdr3_b_aa.apply(lambda x: _valid_cdr3(x))
	# Ensure length CDR3 > 5
	ls = bulk_df.cdr3_b_aa.apply(lambda x: len(x) > 5)
	bulk_df = bulk_df[(v) & (ls)]

	# Select only important columns
	bulk_df = bulk_df[['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'templates', 'productive_frequency','count']]
	# Assign a cid for tracking purposes
	bulk_df['cid'] = [f"cid{i}" for i in bulk_df.index]
	# Load clean bulk data to assign CDR1,2,3
	tr_bulk = TCRrep(	cell_df = bulk_df, 
						organism = 'human', 
						chains = ['beta'], 
						compute_distances= False)

	# Here we expect comma or tab, but we check file for correct seperator (TAB, TWO MANY COMMAS CAUSES A FAILURE)
	#sep_search_filename = check_tsv_csv(
	#	filename = search_filename, 
	#	check_column_names = ['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'max_radi' ,'pgen','regex'])
	
		# Read search file
	search_df = pd.read_csv(search_filename, sep = sep_search_filename )
	# Assign a count column of 1, for purposes of loading data into TCRrep
	search_df['count'] = 1

	# If source and index columns are missing we provide them, select only the relevant columns
	try:
		search_df = search_df[['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'count', 'max_radi' ,'pgen','index','source','regex','target_hits', 'target_seqs']].copy()
	except KeyError:
		search_df['source'] = search_filename
		search_df['index'] = search_df.index.to_list()
		search_df = search_df[['cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'count', 'max_radi' ,'pgen','index','source','regex','target_hits', 'target_seqs']].copy()

	# Load the search file into TCRrep instance, getting CDR1, 2, 2.5 by vgene name infrerence
	tr_search = TCRrep( cell_df = search_df.copy(),
						organism = 'human', 
						chains = ['beta'], 
						compute_distances= False)

	# Compute Rectangular distance (search on rows, bulk on columns)
	tr_search.compute_rect_distances(df = tr_search.clone_df, 
									 df2 = tr_bulk.clone_df, 
									 store = False)

	# Each search sequence has a max radius 
	max_radi = tr_search.clone_df.max_radi.values
	# Convert to a 2D array to permit broadcasting (n,1) with (n,m) rw_matrix
	max_radi = max_radi.reshape(len(max_radi), 1)

	# Assert that [n,1] aligns with [n,m] 
	assert max_radi.shape[0] == tr_search.rw_beta.shape[0]
	# Get index of True, where 
	ij = tr_search.rw_beta < max_radi
	# Get the column index of each bulk sequences within the row specific radius
	icol = [np.where(x) for x in ij]
	bulk_hits = np.sum(ij, axis = 1)
	assert np.all(bulk_hits == [len(x[0]) for x in icol])
	# Retrieve the distance of sequences with the max radius for each row
	idist =[tr_search.rw_beta[i,j] for i,j in enumerate(icol)]

	# Retrieve sequences from the bulk clone_df
	iseqs = [tr_bulk.clone_df['cdr3_b_aa'].iloc[x].to_list() for x in icol]
	ivgenes = [tr_bulk.clone_df['v_b_gene'].iloc[x].to_list() for x in icol]
	ijgenes = [tr_bulk.clone_df['j_b_gene'].iloc[x].to_list() for x in icol]
	
	# Retrieve abundances from the bulk clone df
	itemplates = [tr_bulk.clone_df['templates'].iloc[x].to_list()  for x in icol]
	icounts = [tr_bulk.clone_df['count'].iloc[x].to_list() for x in icol]
	ifreqs = [tr_bulk.clone_df['productive_frequency'].iloc[x].to_list()  for x in icol]

	assert [np.sum(x) for x in itemplates] == [np.sum(x) for x in icounts]
	isumtemplates = [np.sum(x) for x in itemplates]
	isumcounts    = [np.sum(x) for x in icounts]
	isumfreqs     = [np.sum(x) for x in ifreqs ]

	df_summ = pd.DataFrame({ 'bulk_sum_freq'     : isumfreqs,
							 'bulk_sum_templates': isumtemplates,
							 'bulk_sum_counts'   : isumcounts,
							 'bulk_seqs'      	 : iseqs,
							 'bulk_v_genes'   	 : ivgenes,
							 'bulk_j_genes'   	 : ijgenes,
							 'bulk_distances' 	 : idist,
							 'bulk_templates' 	 : itemplates,
							 'bulk_counts'    	 : icounts, 
							 'bulk_freqs'     	 : ifreqs})

	assert tr_search.clone_df.shape[0] == search_df.shape[0]

	#search_df = pd.read_csv(search_filename, sep = sep)
	###### <<>>>>>
	result_df = pd.concat([tr_search.clone_df, df_summ], axis = 1)
	result_df['sourcefile'] = bulk_filename
	result_df['searchfile'] = search_filename
	
	regex_on_bulk = [[re.search(pattern = r['regex'], string =  s) for s in r['bulk_seqs']] for _,r in result_df.iterrows()]
	bulk_regex_match = [[True if (x is not None) else False for x in sublist] for sublist in regex_on_bulk]
	result_df['bulk_regex_match'] = bulk_regex_match
	result_df['bulk_sum_freqs_regex_adj'] = [pd.Series(r['bulk_freqs'], dtype = "float64")[pd.Series(r['bulk_regex_match'], dtype = 'bool')].sum() for i,r in result_df.iterrows()]
	result_df['bulk_sum_templates_regex_adj'] = [pd.Series(r['bulk_templates'], dtype = "int32")[pd.Series(r['bulk_regex_match'], dtype = 'bool')].sum() for i,r in result_df.iterrows()]
	result_df['bulk_sum_counts_regex_adj'] = [pd.Series(r['bulk_counts'], dtype = "int32")[pd.Series(r['bulk_regex_match'],  dtype = 'bool')].sum() for i,r in result_df.iterrows()]
	
	# Tabulating Tcrdist == 0
	ij = tr_search.rw_beta == 0
	icol = [np.where(x) for x in ij]
	itemplates = [tr_bulk.clone_df['templates'].iloc[x].to_list()  for x in icol]
	icounts = [tr_bulk.clone_df['count'].iloc[x].to_list() for x in icol]
	ifreqs = [tr_bulk.clone_df['productive_frequency'].iloc[x].to_list()  for x in icol]
	assert [np.sum(x) for x in itemplates] == [np.sum(x) for x in icounts]
	isumcounts0 = [np.sum(x) for x in icounts]
	isumtemplates0 = [np.sum(x) for x in itemplates]
	isumfreqs0     = [np.sum(x) for x in ifreqs ]
	result_df['bulk_sum_freqs_tcrdist0'] = isumfreqs0
	result_df['bulk_sum_templates_tcrdist0'] = isumtemplates0
	result_df['bulk_sum_counts_tcrdist0'] = isumcounts0


	# Tabulate Perfect Match 
	result_df['cdr3_exact_match'] = [[s == r['cdr3_b_aa'] for s in r['bulk_seqs']] for _,r in result_df.iterrows()]
	result_df['v_gene_exact_match'] = [[s == r['v_b_gene'] for s in r['bulk_v_genes']] for _,r in result_df.iterrows()]
	def safely_compare_boolan_lists(a,b):
		if len(a) < 1:
			r = list()
		else:
			r = np.array(a) & np.array(b)
			r = list(r)
		return r
	result_df['vcdr3_exact_match'] = [safely_compare_boolan_lists(r['cdr3_exact_match'], r['v_gene_exact_match']) for _,r in result_df.iterrows()]
	# Summarize Exact Matches
	result_df['bulk_sum_freqs_vcdr3match'] = [np.sum(pd.Series(r['bulk_freqs'],dtype = 'float64')[r['vcdr3_exact_match']]) for _,r in result_df.iterrows()]
	result_df['bulk_sum_templates_vcdr3match'] = [np.sum(pd.Series(r['bulk_templates'],dtype = 'int32')[r['vcdr3_exact_match']]) for _,r in result_df.iterrows()]
	result_df['bulk_sum_counts_vcdr3match'] = [np.sum(pd.Series(r['bulk_counts'],dtype = 'int32')[r['vcdr3_exact_match']]) for _,r in result_df.iterrows()]

	# Tabulate Perfect Matches to any of the Target Seqs
	result_df['target_seqs'] = [ast.literal_eval(x) for x in result_df['target_seqs']]
	result_df['bulk_seq_within_targetseqs'] = [[s in r['target_seqs'] for s in r['bulk_seqs']] for _,r in result_df.iterrows()]
	result_df['bulk_sum_freqs_within_targetset'] = [np.sum(pd.Series(r['bulk_freqs'],dtype = 'float64')[r['bulk_seq_within_targetseqs']]) for _,r in result_df.iterrows()]
	result_df['bulk_sum_templates_within_targetset'] = [np.sum(pd.Series(r['bulk_templates'],dtype = 'int32')[r['bulk_seq_within_targetseqs']]) for _,r in result_df.iterrows()]
	result_df['bulk_sum_counts_within_targetset'] = [np.sum(pd.Series(r['bulk_counts'],dtype = 'int32')[r['bulk_seq_within_targetseqs']]) for _,r in result_df.iterrows()]


	desired_output_column_order = [  'cdr3_b_aa',
									 'v_b_gene',
									 'j_b_gene',
									 'max_radi',
									 'pgen',
									 'index',
									 'source',
									 'target_seqs',
									 'regex',
									 'cdr1_b_aa',
									 'cdr2_b_aa',
									 'pmhc_b_aa',
									 'count',
									 'clone_id',
									 'sourcefile',
									 'searchfile',
									 
									 'bulk_distances',
									 'bulk_templates',
									 'bulk_counts',
									 'bulk_freqs',
									 'bulk_seqs',
									 'bulk_v_genes',
									 'bulk_j_genes',
									 'bulk_regex_match',

									 'bulk_sum_freq',
									 'bulk_sum_templates',
									 'bulk_sum_counts',
									  
									 'bulk_sum_freqs_regex_adj',
									 'bulk_sum_templates_regex_adj',
									 'bulk_sum_counts_regex_adj',
									 
									 'bulk_sum_freqs_tcrdist0',
									 'bulk_sum_templates_tcrdist0',
									 'bulk_sum_counts_tcrdist0',
									 #'cdr3_exact_match',
									 #'v_gene_exact_match',
									 #'vcdr3_exact_match',
									 'bulk_sum_freqs_vcdr3match',
									 'bulk_sum_templates_vcdr3match',
									 'bulk_sum_counts_vcdr3match',

									 'bulk_sum_freqs_within_targetset',
									 'bulk_sum_templates_within_targetset',
									 'bulk_sum_counts_within_targetset']

	result_df = result_df[desired_output_column_order] 

	# COMPRESS SPARSE MATRIX FOR LATER REFRENCE
	tr_search.rw_beta[tr_search.rw_beta == 0]  = 1 # set true zeros to 1
	tr_search.rw_beta[tr_search.rw_beta > 50]  = 0 # ignores everything less than 100
	rw_beta_sparse = scipy.sparse.csr_matrix(tr_search.rw_beta)
	tr_search.rw_beta = rw_beta_sparse

	return result_df, rw_beta_sparse, tr_search



from tcrdist.public import _neighbors_sparse_variable_radius, _neighbors_sparse_fixed_radius
import pandas as pd

def join_by_dist(
    csrmat,
    left_df,
    right_df,
    how = "inner",
    left_cols  = ['cdr3_b_aa','v_b_gene','j_b_gene','subject'],
    right_cols = ['cdr3_b_aa','v_b_gene','j_b_gene','subject'],
    left_suffix = '_x',
    right_suffix = '_y',
    max_n= 5,
    radius = 1,
    radius_list = None, 
    sort_by_dist = True):
    """
    Join two sets of TCRs, based on a distance threshold 
    encoded in a sparse matrix `csrmat`. 
    
    The TCRs in the Left-DataFrame, are joined with TCRs in the Right-Dataframe, 
    for up to `max_n` closest TCRs where the paired distance is less that 
    that specifed in the `radius` or `radius_list` arguments.
    
    This is analogous to SQL type joins, except instead of matching common keys, the 
    rows of the Left and Right DataFrames are merged based on finding simimlar TCRs within
    a specified radius of sequence divergence. Intutively, this 
    permits fuzzy matching between similar TCRs in any two 
    TCR repertoire data sets.
    
    Crucially, one must provide a scipy.sparse csr matrix which can be pre-computed using. 
    :py:func:`tcrdist.rep_funcs.compute_pws_sparse` or 
    :py:func:`tcrdist.reperotire.TCRrep.compute_sparse_rect_distances`
    
    It is also possible to join using a unique radius for each sequence in Left-DataFrame
    using the `radius_list` argument instead of the fixed `radius` argument. 
    However, if using a radius list, it must match the number of rows in the csrmat 
    and the number of rows in Left DataFrame (i.e., len(radius_list) == left_df.shape[1] ).
    
    Parameters 
    ----------
    
    csrmat : scipy.sparse.matrix
        rows must correspond to index of left_df
        columns must correspond to index of right_df
    left_df: pandas.DataFrame
        Clone DataFrame
    right_df: pandas.DataFrame
        Clone DataFrame
    how : str
        Must be one of 'inner','left', or'outer', which determines the type of merge to be performed.

        * 'inner' use intersection of top max_n neigbors between Left and Right DataFrames, droping rows where there is no match.
        
        * 'left' use top max_n rows from Right DataFrame neighboring rows in from Left DataFrame or produce NA where a TCR in Left DataFrame has no neighbor in the Right DataFrame.
 
        * 'outer' a FULL OUTER JOIN combines top max_n neighbors of both left and right DataFrame, producing NAs where a TCR in either the Right or Left DataFrame has no neighbor in other DataFrame.

        * (hint: right joins are not possible, unless you switch input dataframe order and recompute the spase matrix)
        
    left_cols : list
        all columns to include in left_df
    right_cols : list
        all columns to include in right_df 
    left_suffix : str
        appends to left columns
    right_suffix : str
        appends to right columns
    max_neighors : int
        limit on number of neighbors to join per row. 
        For instance if a TCR has 100 neighbors only the first 10 rows in the right df will
        be included (this is to avoid massive blowup in cases where knowing 
        about a few neighbors would suffice)
    radius = int
        default is 1
    
    Returns
    -------
    left_right_df : pandas DataFrame
        concatenates rows from left and right dataframe for all sequences within a specified distance
    """
    assert how in ['inner','left','outer']
    if how == "inner":
        add_unmatched_left = False
        add_unmatched_right= False
    elif how == "left":
        add_unmatched_left = True
        add_unmatched_right= False
    elif how == "outer":
        add_unmatched_left = True
        add_unmatched_right= True
    
    if radius_list is None:
        nn = _neighbors_sparse_fixed_radius(csrmat = csrmat, radius = radius)
    else: 
        assert len(radius_list) == left_df.shape[0]
        nn = _neighbors_sparse_variable_radius(csrmat = csrmat, radius_list = radius_list)
    left_index = list()
    right_index = list()
    dists = list()

    for i,ns in enumerate(nn):
        l = len(ns)
        if l > 0:
            ds = csrmat[i,ns].data
            if sort_by_dist:
                # Sort n index by dist smallest to largest
                # sorted(zip([10,1,0,-1],[100,500,1,10000])) => [(-1, 10000), (0, 1), (1, 500), (10, 100)]
                # thus [n for d, n in sorted(zip([10,1,0,-1],[100,500,1,10000]))] => [10000, 1, 500, 100]
                ns_ds = [(n,d) for d, n in sorted(zip(ds, ns))]
                ns,ds = zip(*ns_ds)
            if l > max_n:
                l = max_n
            left_index.extend([i]*l)
            right_index.extend(ns[0:l])
            dists.extend(ds[0:l])

    left_selection        = left_df[left_cols].rename(columns ={k:f"{k}{left_suffix}" for k in left_cols}).iloc[left_index,].reset_index(drop = True)
    right_selection       = right_df[right_cols].rename(columns ={k:f"{k}{right_suffix}" for k in right_cols}).iloc[right_index,].reset_index(drop = True)
    left_right_df         = pd.concat([left_selection, right_selection], axis = 1) 
    left_right_df['dist'] = dists
    
    
    if add_unmatched_left:
        left_index_unmatched = sorted(list(set(left_df.index) - set(left_index)))
        left_df_unmatched = left_df[left_cols].rename(columns ={k:f"{k}{left_suffix}" for k in left_cols}).iloc[left_index_unmatched,].reset_index(drop = True)
        left_right_df = pd.concat([left_right_df,left_df_unmatched],axis= 0)
    if add_unmatched_right:
        right_index_unmatched = sorted(list(set(right_df.index) - set(right_index)))
        right_df_unmatched = right_df[right_cols].rename(columns ={k:f"{k}{right_suffix}" for k in right_cols}).iloc[right_index_unmatched,].reset_index(drop = True)
        left_right_df = pd.concat([left_right_df,right_df_unmatched],axis = 0)
        
    return left_right_df
  

 



import pytest 



def test_introduction_3():
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = False)

    from tcrdist.plotting import plot_pairings, _write_svg
   
    svg_PA  = plot_pairings(tr.clone_df.loc[tr.clone_df.epitope == "PA"], 
                cols = ['v_b_gene', 'j_b_gene','j_a_gene', 'v_a_gene'], 
                count_col='count')
      
    svg_NP = plot_pairings(tr.clone_df.loc[tr.clone_df.epitope == "NP"], 
            cols = ['v_b_gene', 'j_b_gene', 'j_a_gene', 'v_a_gene'], 
            count_col='count')
    
    _write_svg(svg_PA, name = "PA_gene_usage_plot.svg", dest = ".")
    
    _write_svg(svg_NP, name = "NP_gene_usage_plot.svg", dest = ".")

    import fishersapi
    fishersapi.fishers_frame(tr.clone_df.loc[tr.clone_df.epitope == "NP"], 
                             col_pairs=[('v_b_gene', 'j_b_gene'),
                                        ('v_a_gene', 'j_a_gene'),
                                        ('v_a_gene', 'v_b_gene'),
                                        ('j_a_gene', 'j_b_gene')])


import pytest
import Levenshtein
def my_own_metric(s1,s2):   
    return Levenshtein.distance(s1,s2)
def test_example_5():
    """
    If you want a tcrdistance, but you want to use your own metric. 
    (A valid metric takes two strings and returns a numerical distance).  

    def my_own_metric(s1,s2):   
        return Levenshtein.distance(s1,s2)
    """
    import pwseqdist as pw
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    import multiprocessing

    df = pd.read_csv("dash.csv")
    df = df.head(100) # for faster testing
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                use_defaults=False,
                compute_distances = False,
                cpus = 1,
                db_file = 'alphabeta_gammadelta_db.tsv')

    metrics_a = {
        "cdr3_a_aa" : my_own_metric ,
        "pmhc_a_aa" : my_own_metric ,
        "cdr2_a_aa" : my_own_metric ,
        "cdr1_a_aa" : my_own_metric }

    metrics_b = {
        "cdr3_b_aa" : my_own_metric ,
        "pmhc_b_aa" : my_own_metric ,
        "cdr2_b_aa" : my_own_metric,
        "cdr1_b_aa" : my_own_metric }

    weights_a = { 
        "cdr3_a_aa" : 1,
        "pmhc_a_aa" : 1,
        "cdr2_a_aa" : 1,
        "cdr1_a_aa" : 1}

    weights_b = { 
        "cdr3_b_aa" : 1,
        "pmhc_b_aa" : 1,
        "cdr2_b_aa" : 1,
        "cdr1_b_aa" : 1}

    kargs_a = {  
        'cdr3_a_aa' : 
            {'use_numba': False},
        'pmhc_a_aa' : {
            'use_numba': False},
        'cdr2_a_aa' : {
            'use_numba': False},
        'cdr1_a_aa' : {
            'use_numba': False}
        }
    kargs_b = {  
        'cdr3_b_aa' : 
            {'use_numba': False},
        'pmhc_b_aa' : {
            'use_numba': False},
        'cdr2_b_aa' : {
            'use_numba': False},
        'cdr1_b_aa' : {
            'use_numba': False}
        }

    tr.metrics_a = metrics_a
    tr.metrics_b = metrics_b

    tr.weights_a = weights_a
    tr.weights_b = weights_b

    tr.kargs_a = kargs_a 
    tr.kargs_b = kargs_b

    tr.compute_distances()

    tr.pw_cdr3_b_aa
    tr.pw_beta


"""
test_public.py
"""
import pandas as pd
import numpy as np
import os
import pytest 
from tcrdist.repertoire import TCRrep
from tcrsampler.sampler import TCRsampler
from tcrdist.public import *
from tcrdist.tree import _default_tcrsampler_mouse_beta, _default_tcrsampler_mouse_alpha
import scipy

# Fixture Mouse Beta
df = pd.read_csv("dash.csv").query('epitope == "PA"').reset_index().copy()
tr = TCRrep(cell_df = df.head(300).copy(), 
            organism = 'mouse', 
            chains = ['beta'], 
            db_file = 'alphabeta_gammadelta_db.tsv', 
            compute_distances = True)
tr.clone_df['radius'] = 23
tsampler_beta = _default_tcrsampler_mouse_beta()

# Fixture Mouse Alpha
tra = TCRrep(cell_df = df.head(300).copy(), 
            organism = 'mouse', 
            chains = ['alpha'], 
            db_file = 'alphabeta_gammadelta_db.tsv', 
            compute_distances = True)
tra.clone_df['radius'] = 23
tsampler_alpha = _default_tcrsampler_mouse_alpha()


def test_TCRpublic():
	df = pd.read_csv("dash.csv").query('epitope == "PA"').reset_index().copy()
	tr = TCRrep(cell_df = df.head(200).copy(), 
	            organism = 'mouse', 
	            chains = ['beta'], 
	            db_file = 'alphabeta_gammadelta_db.tsv', 
	            compute_distances = True)
	tr.clone_df['radius'] = 40
	from tcrdist.public import TCRpublic
	tp = TCRpublic(tcrrep = tr, organism = 'mouse', chain = 'beta')
	tp.report()

def test__neighbors_fixed_radius():
	""" Returns the list of neighbor column indices if within the fixed radius """
	result = _neighbors_fixed_radius(pwmat = tr.pw_beta, radius = 20) 
	assert isinstance(result, list)

def test__K_neighbors_fixed_radius():
	result = _K_neighbors_fixed_radius(pwmat = tr.pw_beta, radius = 20) 
	assert isinstance(result, list)

def test__neighbors_variable_radius():
	""" Returns the list of neighbor column indices if within the fixed radius """
	result = _neighbors_variable_radius(pwmat = tr.pw_beta, radius_list = tr.clone_df['radius'].to_list())
	assert isinstance(result, list)

def test__K_neighbors_variable_radius():
	""" Returns the number of neighbors (self-inclusive) if within the fixed radius"""
	result = _K_neighbors_variable_radius(pwmat = tr.pw_beta, radius_list = tr.clone_df['radius'].to_list())
	assert isinstance(result, list)

def test_neighbors_sparse_fixed_radius():
	M = np.array([[-1,0,1,4],[1,-1,0,2],[0,10,-1,3],[0,0,2,-1]])
	S = scipy.sparse.csr_matrix(M)
	NN = _neighbors_sparse_fixed_radius(csrmat = S, radius = 4)
	assert NN == [[0, 2, 3], [0, 1, 3], [2, 3], [2, 3]]

def test_neighbors_sparse_variable_radius():
	M = np.array([[-1,1,1,4],[1,-1,1,2],[1,10,-1,3],[5,1,2,-1]])
	S = scipy.sparse.csr_matrix(M)
	N = _neighbors_sparse_variable_radius(csrmat = S, radius_list = [1,1,20,1])
	assert N ==  [[0, 1, 2], [0, 1, 2], [0, 1, 2, 3], [1,3]]

def test_make_motif_logo():
	
	svg, svg_raw = make_motif_logo(
			tcrsampler = tsampler_beta,
			clone_df = tr.clone_df,
			pwmat = tr.pw_beta,
			centroid = 'CASSDFDWGGDAETLYF',
			v_gene = 'TRBV13-1*01',
			radius = 24,
			pwmat_str = 'pw_beta',
			cdr3_name = 'cdr3_b_aa',
			v_name = 'v_b_gene',
			gene_names = ['v_b_gene','j_b_gene'])
	assert isinstance(svg, str)
	assert svg.startswith('<svg')
	assert isinstance(svg_raw, str)
	assert svg_raw.startswith('<svg')

# TODO: Test with alpha,beta,gamma,delta # Test with nndif and hdif
def test_quasi_public_meta_clonotypes_beta():
	RB = _quasi_public_meta_clonotypes(clone_df = tr.clone_df.copy(), 
								  pwmat = tr.pw_beta.copy(),
								  tcrsampler = tsampler_beta, 
								  cdr3_name = 'cdr3_b_aa',
								  v_gene_name = 'v_b_gene',
								  nr_filter = True,
								  output_html_name = "quasi_public_clones_beta_testonly.html",
								  labels = ['clone_id',
								  			'cdr3_b_aa', 
								  			'v_b_gene',
								  			'j_b_gene',
								  			'radius',
								  			'neighbors',
								  			'K_neighbors',
								  			#'cdr3s',
								  			'nsubject',
								  			'qpublic',
								  			'cdr3_b_aa.summary',
								  			'v_b_gene.summary',
								  			'j_b_gene.summary',
								  			'subject.summary'],
								  fixed_radius = False,
								  radius = 24,
								  query_str = 'qpublic == True & K_neighbors > 3',
								  kargs_member_summ = {
								 	'key_col'   : 'neighbors', 
								 	'count_col' : 'count',
								 	'addl_cols' : ['subject'],
								 	'addl_n'    : 4},
								  kargs_motif = {
								  	'pwmat_str'  : 'pw_beta',
									'cdr3_name'  : 'cdr3_b_aa',
									'v_name'     : 'v_b_gene',
									'gene_names' : ['v_b_gene','j_b_gene']})



def test_quasi_public_meta_clonotypes_beta_from_nieghbor_diff():
	from tcrdist.rep_diff import neighborhood_diff
	ndif = neighborhood_diff(   clone_df= tr.clone_df, 
								pwmat = tr.pw_beta, 
								count_col = 'count', 
								x_cols = ['epitope'], 
								knn_radius = 25)

	nn_clone_df = pd.concat([tr.clone_df, ndif[['neighbors', 'K_neighbors','val_0','ct_0']] ], axis = 1)
	RB = _quasi_public_meta_clonotypes(clone_df = nn_clone_df, 
							  pwmat = tr.pw_beta.copy(),
							  tcrsampler = tsampler_beta, 
							  cdr3_name = 'cdr3_b_aa',
							  v_gene_name = 'v_b_gene',
							  nr_filter = True,
							  output_html_name = "quasi_public_clones_beta_testonly_from_NN.html",
							  labels = ['clone_id',
							  			'val_0','ct_0', # <- NNDIF COLUMNS
							  			'cdr3_b_aa', 
							  			'v_b_gene',
							  			'j_b_gene',
							  			'radius',
							  			'neighbors',
							  			'K_neighbors',
							  			'nsubject',
							  			'qpublic',
							  			'cdr3_b_aa.summary',
							  			'v_b_gene.summary',
							  			'j_b_gene.summary',
							  			'subject.summary'],
							  fixed_radius = False,
							  radius = 24,
							  query_str = 'qpublic == True & K_neighbors > 3',
							  kargs_member_summ = {
							 	'key_col'   : 'neighbors', 
							 	'count_col' : 'count',
							 	'addl_cols' : ['subject'],
							 	'addl_n'    : 4},
							  kargs_motif = {
							  	'pwmat_str'  : 'pw_beta',
								'cdr3_name'  : 'cdr3_b_aa',
								'v_name'     : 'v_b_gene',
								'gene_names' : ['v_b_gene','j_b_gene']})



def test_quasi_public_meta_clonotypes_alpha():
	RA = _quasi_public_meta_clonotypes(clone_df = tra.clone_df.copy(), 
							  pwmat = tra.pw_alpha.copy(),
							  tcrsampler = tsampler_alpha, 
							  cdr3_name = 'cdr3_a_aa',
							  v_gene_name = 'v_a_gene',
							  nr_filter = True,
							  output_html_name = "quasi_public_clones_alpha_testonly.html",
							  labels = ['clone_id',
							  			'cdr3_a_aa', 
							  			'v_a_gene',
							  			'j_a_gene',
							  			'radius',
							  			'neighbors',
							  			'K_neighbors',
							  			#'cdr3s',
							  			'nsubject',
							  			'qpublic',
							  			'cdr3_a_aa.summary',
							  			'v_a_gene.summary',
							  			'j_a_gene.summary',
							  			'subject.summary'],
							  fixed_radius = False,
							  radius = 24,
							  query_str = 'qpublic == True & K_neighbors > 3',
							  kargs_member_summ = {
							 	'key_col'   : 'neighbors', 
							 	'count_col' : 'count',
							 	'addl_cols' : ['subject'],
							 	'addl_n'    : 4},
							  kargs_motif = {
							  	'pwmat_str'  : 'pw_alpha',
								'cdr3_name'  : 'cdr3_a_aa',
								'v_name'     : 'v_a_gene',
								'gene_names' : ['v_a_gene','j_a_gene']})
	
		




from zipfile import ZipFile
import requests
from tcrdist import paths

__all__ = ['download_and_extract_zip_file']

"""
python -c "from tcrdist.setup_tests import *; download_and_extract_zip_file('bulk.csv.zip')"
python -c "from tcrdist.setup_tests import *; download_and_extract_zip_file('dash.zip')"
python -c "from tcrdist.setup_tests import *; download_and_extract_zip_file('sant.csv.zip')"
"""

# public facing data files, L looksup url based for dropbox and aws
L = {"dash.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/pce3f9816ntzjki/dash.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "bulk.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/4yy9110al33ckh7/bulk.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "olga_T_alpha_beta_1000K_simulated_cdr3.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/6qcxs3ylmczyfk7/olga_T_alpha_beta_1000K_simulated_cdr3.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "cdr3_beta_500K.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/yevk0rus1dqnzcg/cdr3_beta_500K.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "human_T_alpha_beta_sim200K.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/jjnon2x8qt0qk4y/human_T_alpha_beta_sim200K.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "vdjDB_PMID28636592.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/mmjyi8i3p1ps3qq/vdjDB_PMID28636592.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "sant.csv.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/8p3djrdd270ad0n/sant.csv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "bulk.csv.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/g6k2h1ed5d5sabz/bulk.csv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "wiraninha_sampler.zip":
        {'dropbox': {
            'url' : "https://www.dropbox.com/s/ily0td3tn1uc7bi/wiraninha_sampler.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "ruggiero_mouse_sampler.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/yz8v1c1gf2eyzxk/ruggiero_mouse_sampler.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "ruggiero_human_sampler.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/jda6qtemk65zlfk/ruggiero_human_sampler.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "britanova_human_beta_t_cb.tsv.sampler.tsv.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/87n5v2by80xhy1q/britanova_human_beta_t_cb.tsv.sampler.tsv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "emerson_human_beta_t_cmvneg.tsv.sampler.tsv.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/04mxrzw7f5wkg1x/emerson_human_beta_t_cmvneg.tsv.sampler.tsv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "ruggiero_human_alpha_t.tsv.sampler.tsv.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/9h84bzhd0asfym7/ruggiero_human_alpha_t.tsv.sampler.tsv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "ruggiero_human_beta_t.tsv.sampler.tsv.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/onr5lntmlm4fivi/ruggiero_human_beta_t.tsv.sampler.tsv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "ImmunoSeq_MIRA_matched_tcrdist3_ready.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/1vma8opj0yqts9e/ImmunoSeq_MIRA_matched_tcrdist3_ready.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    "ImmunoSeq_MIRA_matched_tcrdist3_ready_2_files.zip":
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/qrjawanmrklts70/ImmunoSeq_MIRA_matched_tcrdist3_ready_2_files.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    'bioRxiv_v2_metaclonotypes.tsv.zip':
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/hpt1ropv7u02eqr/bioRxiv_v2_metaclonotypes.tsv.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    'ImmunoSEQhsTCRBV4b_tcrdist3.zip':
        {'dropbox':{
            'url' : "https://www.dropbox.com/s/22iyel9uyzy7zyq/ImmunoSEQhsTCRBV4b_tcrdist3.zip?dl=1"},
        'aws': { 
            'url' : None}
    },
    '2021-04-02-Release_v2.1_metaclonotypes_concise.tsv.zip':
        {'dropbox':{
         'url': "https://www.dropbox.com/s/6to8fsga8k5twdr/2021-04-02-Release_v2.1_metaclonotypes_concise.tsv.zip?dl=1"},
         'aws':{
            'url' : None}
    },
    '2021-04-02-Release_v2.1_TCRs_concise_covid_only.tsv.zip':
        {'dropbox' : {
            'url' : "https://www.dropbox.com/s/60i0reiv7utr8hw/2021-04-02-Release_v2.1_TCRs_concise_covid_only.tsv.zip?dl=1"},
         'aws':{
            'url' : None}
    }
    }
    

def list_available_zip_files():
    """
    List all available zip files downloadable from tcrdist3. 
    
    Returns 
    -------
    List of zipfile names that can be passed to zipfile argument in download_and_extract_zip_file()
    """

    return [k for k in L.keys()]

def get_url(zipfile, source = 'dropbox'):
    """
    Lookup the url associatd with a zipfile

    Returns
    -------
    url : str
    """
    url = L[zipfile][source]['url']
    return url

def download_and_extract_zip_file(zipfile = None, source = 'dropbox', dest = paths.path_to_base):
    """ 
    Downloads and extracts a zip file to destination folder.
    Uses functions from **requests** and **Zipfile**, part of the Python Standard Library, to avoid the 
    platform independent use of wget, curl, gunzip, etc.

    
    Parameters 
    ----------
    zipfile : str
        Name of zip file see (list_available_zip_files() for current list)
    source : str
        The host source name where the file will be downloaded from. Currently 'dropbox' 
        is the only aviable option but 'aws' will be available on release >= 1.0.0
    dest : str
        path where the files to be saved and unzipped
    
    """
    url = get_url(zipfile, source = source)
    r = requests.get(url) 
    with open(zipfile,'wb') as f:
        f.write(r.content)
    with ZipFile(zipfile, 'r') as zipObj:
        # Extract all the contents of zip file in different directory
        zipObj.extractall(dest)

def download_and_extract_directly_from_url(zipfile, url,dest = paths.path_to_base):
    """
    Advanced feature allowing downloads from URLs not prespecified in list above.
    
    Users who want to add downloads to public URLs may find this useful; However, 
    this is not recommended unless you want something not yet available in package test files. 
    You can check list_available_zip_files() to see if the name and url of the file 
    you want is already prespecified as a test or demonstration file. 
    
    Whereas if you use this direct option, filenames are not strictly enforced.
    
    Parameters
    ----------
    zipfile : str 
        Name of zip file
    url : str
        User can directly provide the url they wish to use (advanced option)
    dest : str
        path where the files to be saved and unzipped
    
    Example
    -------
    from tcrdist.setup_tests import download_and_extract_directly_from_url
    download_and_extract_directly_from_url(zipfile= "2021-04-02-Release_v2.1_TCRs_concise_covid_only.tsv.zip",
        url = "https://www.dropbox.com/s/60i0reiv7utr8hw/2021-04-02-Release_v2.1_TCRs_concise_covid_only.tsv.zip?dl=1")
    
    """
    r = requests.get(url) 
    with open(zipfile,'wb') as f:
        f.write(r.content)
    with ZipFile(zipfile, 'r') as zipObj:
        # Extract all the contents of zip file in different directory
        zipObj.extractall(dest)
    



"""
swap_gene_names.py 

2020-07-31
Updated 2020-10-29: moved mapping to a CSV in the db folder.

Adaptive MIRA Release 2 to tcrdist3 alphabeta_gammadelta_db.tsv
Generated 2020-07-31 Based on universal.py 
"""
import pandas as pd
from os.path import join as opj

from .paths import path_to_db

__all__ = ['adaptive_to_imgt']

mapping = pd.read_csv(opj(path_to_db, 'adaptive_imgt_mapping.csv'))

hum = mapping.loc[mapping['species'] == 'human'].set_index('adaptive')['imgt'].fillna('NA').to_dict()
mus = mapping.loc[mapping['species'] == 'mouse'].set_index('adaptive')['imgt'].fillna('NA').to_dict()

adaptive_to_imgt = {'human':hum,
                    'mouse':mus}

# This quickly makes a dictionary from  'TRGV7*00' -> 'TRGV7*01'
vdj00_to_imgt = dict()
vdj00_to_imgt['human']=\
	{f"{v.split('*')[0]}":v for k,v in adaptive_to_imgt['human'].items() }
vdj00_to_imgt['mouse']=\
	{f"{v.split('*')[0]}":v for k,v in adaptive_to_imgt['mouse'].items() }





"""
This is an integration test of the major features in background.py
"""
import os
import pytest

__all__ = ['test_background_generation_in_mira_60'] 

def test_background_generation_toy_example():
	import sys
	import os
	import numpy as np
	import pandas as pd
	from tcrsampler.sampler import TCRsampler
	from tcrdist.background import make_gene_usage_counter, get_gene_frequencies, calculate_adjustment, make_gene_usage_counter
	from tcrdist.background import make_vj_matched_background, make_flat_vj_background
	from tcrdist.background import get_stratified_gene_usage_frequency
	from tcrdist.background import sample_britanova
	"""
	SUPPOSE WE HAVE SOME REPERTOIRE WITH THE FOLLOWING GENE USAGE SPECIFIED BY ix
	< df_target > For testing we will use a set of 25 TCRs generated from rare and semi-rare V,J pairings. We use 25 only 
	because we will be comuting distances against 4.6 Million seqs.
		1. TCRsampler, replacing gene occurance frequencies with subject tratified estimates
		NOTE: with replace = True .vj_occur_freq will now be the stratified value
		2. Make V,J gene usage matched backgound to match usage in df_target
		3. Use a subject-stratifeid random draw from the Britanova Chord Blood Samples
		4. Make V,J gene usage matched backgound to match usage in df_target
	"""
	ts = TCRsampler(default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv') 	# 1
	ts = get_stratified_gene_usage_frequency(ts = ts, replace = True)

	ix =[['TRBV19*01', 'TRBJ2-5*01', 3],['TRBV24-1*01', 'TRBJ2-4*01', 3],['TRBV25-1*01', 'TRBJ2-4*01', 3],['TRBV30*01', 'TRBJ2-3*01', 2],['TRBV5-4*01', 'TRBJ2-3*01', 2],['TRBV11-2*01', 'TRBJ2-2*01', 2],['TRBV2*01', 'TRBJ1-5*01', 1],['TRBV12-5*01', 'TRBJ2-7*01', 1],['TRBV4-1*01', 'TRBJ1-6*01', 1],['TRBV6-5*01', 'TRBJ1-6*01', 1],['TRBV13*01', 'TRBJ2-3*01', 1],['TRBV18*01', 'TRBJ2-3*01', 1],['TRBV14*01', 'TRBJ2-7*01', 1],['TRBV6-6*01', 'TRBJ2-7*01', 1],['TRBV10-3*01', 'TRBJ2-3*01', 1],['TRBV7-2*01', 'TRBJ2-1*01', 1],['TRBV5-1*01', 'TRBJ2-1*01', 1]]
	flatten = lambda l: [item for sublist in l for item in sublist]
	df_target = pd.concat([pd.DataFrame({'cdr3_b_aa' : flatten(ts.sample([[x[0], x[1], x[2]]])) , 'v_b_gene':x[0], 'j_b_gene':x[1]}) for x in ix]).reset_index(drop = True)

	gene_usage_counter = make_gene_usage_counter(df_target)  							# 2
	df_vj_bkgd = make_vj_matched_background(ts = ts,
		gene_usage_counter = gene_usage_counter, 	
		size = 101000, # Ask for a few extra as Olga can return none if it makes too many non-productive CDR3s
		recomb_type="VDJ", 
		chain_folder = "human_T_beta",
		cols = ['v_b_gene', 'j_b_gene', 'cdr3_b_aa'])
	df_vj_bkgd = df_vj_bkgd.sample(100000).reset_index(drop = True)
	df_vj_bkgd['weights'] = calculate_adjustment(df = df_vj_bkgd, adjcol = "pVJ")
	df_vj_bkgd['source'] = "vj_matched"

	df_britanova_100K = sample_britanova(size = 100000)									# 3
	df_britanova_100K = get_gene_frequencies(ts = ts, df = df_britanova_100K) 
	df_britanova_100K['weights'] = 1
	df_britanova_100K['source'] = "stratified_random"
	df_bkgd = pd.concat([df_vj_bkgd, df_britanova_100K], axis = 0).\
		reset_index(drop = True) 														# 4 

	assert df_bkgd.shape[0] == 200000
	"""
	Visually inspect the gene_usage between target seqs and vj-matched background
	"""
	df_check_match = pd.concat([df_vj_bkgd.groupby(['v_b_gene', 'j_b_gene']).size()/df_vj_bkgd.shape[0], df_target.groupby(['v_b_gene', 'j_b_gene']).size()/df_target.shape[0]], axis = 1)
	assert np.all(abs(df_check_match[0] - df_check_match[1]) < 0.001)
	return df_bkgd


@pytest.mark.skip(reason="This test documents an example, but is redundant. Skipped in the interest of CI time.")
def test_background_generation_in_mira_60(fn = os.path.join('tcrdist','data','covid19','mira_epitope_60_436_MWSFNPETNI_SFNPETNIL_SMWSFNPET.tcrdist3.csv')):
	import sys
	import os
	import numpy as np
	import pandas as pd
	from tcrsampler.sampler import TCRsampler
	from tcrdist.background import make_gene_usage_counter, get_gene_frequencies, calculate_adjustment, make_gene_usage_counter
	from tcrdist.background import make_vj_matched_background, make_flat_vj_background
	from tcrdist.background import get_stratified_gene_usage_frequency
	from tcrdist.background import sample_britanova
	"""
	SUPPOSE WE HAVE SOME REPERTOIRE WITH THE FOLLOWING GENE USAGE SPECIFIED BY ix
	< df_target > For testing we will use a set of 25 TCRs generated from rare and semi-rare V,J pairings. We use 25 only 
	because we will be comuting distances against 4.6 Million seqs.
		1. TCRsampler, replacing gene occurance frequencies with subject tratified estimates
		NOTE: with replace = True .vj_occur_freq will now be the stratified value
		2. Make V,J gene usage matched backgound to match usage in df_target
		3. Use a subject-stratifeid random draw from the Britanova Chord Blood Samples
		4. Make V,J gene usage matched backgound to match usage in df_target
	"""
	ts = TCRsampler(default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv') 	# 1
	ts = get_stratified_gene_usage_frequency(ts = ts, replace = True)

	df_target = pd.read_csv(fn)
	df_target =	df_target[['v_b_gene','j_b_gene','cdr3_b_aa']]

	gene_usage_counter = make_gene_usage_counter(df_target)  							# 2

	df_vj_bkgd = make_vj_matched_background(ts = ts,
		gene_usage_counter = gene_usage_counter, 	
		size = 150000, # Ask for a few extra as Olga can return none if it makes too many non-productive CDR3s
		recomb_type="VDJ", 
		chain_folder = "human_T_beta",
		cols = ['v_b_gene', 'j_b_gene', 'cdr3_b_aa'])
	df_vj_bkgd = df_vj_bkgd.sample(100000).reset_index(drop = True)
	df_vj_bkgd['weights'] = calculate_adjustment(df = df_vj_bkgd, adjcol = "pVJ")
	df_vj_bkgd['source'] = "vj_matched"

	df_britanova_100K = sample_britanova(size = 100000)									# 3
	df_britanova_100K = get_gene_frequencies(ts = ts, df = df_britanova_100K) 
	df_britanova_100K['weights'] = 1
	df_britanova_100K['source'] = "stratified_random"
	df_bkgd = pd.concat([df_vj_bkgd, df_britanova_100K], axis = 0).\
		reset_index(drop = True) 														# 4 

	assert df_bkgd.shape[0] == 200000
	#df_bkgd.
	return df_bkgd






import pandas as pd 
from tcrdist.repertoire import TCRrep


def test_example_12():
	"""
	Illustrate how to save a reload from and archive file
	"""
	import pandas as pd
	from tcrdist.repertoire import TCRrep
	import numpy as np

	df = pd.read_csv("dash.csv").head(100)
	tr = TCRrep(cell_df = df,
		organism = 'mouse',
		chains = ['alpha','beta'],
		db_file = 'alphabeta_gammadelta_db.tsv',
		store_all_cdr=False,
		archive_result=True,
		archive_name = "example_archive")

	tr2 = TCRrep(cell_df = None,
		organism = 'mouse',
		chains = ['alpha','beta'],
		db_file = 'alphabeta_gammadelta_db.tsv',
		blank = True,
		archive_name = "example_archive")
	tr2.rebuild()

	# Check that all atrributes are the same after rebuild, except metrics which can't be zipped
	for k in tr2.__dict__.keys():
		print(k)
		if k in ['all_genes','metrics_a','metrics_b','metrics_d', 'metrics_g',
				'kargs_a','kargs_b','kargs_d','kargs_g']:
			pass
		else:
			assert np.all(getattr(tr, k) == getattr(tr2, k) )

	for k in ['all_genes','metrics_a','metrics_b', 'kargs_a','kargs_b']:
		assert isinstance(getattr(tr2, k), dict)
		assert set(getattr(tr, k).keys()) - set(getattr(tr2, k).keys()) == set()


import pytest 

def test_example_10():
    """
    If you just want a 'tcrdistances' of some target seqs against another set.

    (1) cell_df is asigned the first 10 cells in dash.csv
    (2) TCRrep with default settings.
    (3) compute rectangular distance between clone_df and df2.
    """
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv").head(10)   #(1)
    tr = TCRrep(cell_df = df,               #(2)
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                compute_distances = False)
    df2 = pd.read_csv("dash.csv")
    tr2 = TCRrep(cell_df = df2,               #(2)
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                compute_distances = False)
  
    tr.compute_rect_distances( df = tr.clone_df, df2 = tr2.clone_df) #(3)

    assert tr.rw_alpha.shape == (10,1920) 
    assert tr.rw_beta.shape  == (10,1920)



import pytest 



def test_example_rhesus():
    """
    If you just want a 'tcrdistances' using pre-set default setting.
    
        You can access distance matrices:
            tr.pw_alpha     - alpha chain pairwise distance matrix
            tr.pw_beta      - alpha chain pairwise distance matrix
            tr.pw_cdr3_a_aa - cdr3 alpha chain distance matrix
            tr.pw_cdr3_b_aa - cdr3 beta chain distance matrix
    """
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("rhesus.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'rhesus', 
                chains = ['alpha','beta'], 
                db_file = 'combo_xcr_2023-12-30.tsv')
    
    tr.pw_alpha
    tr.pw_beta
    tr.pw_cdr3_a_aa
    tr.pw_cdr3_b_aa





import pytest
"""
Unit Tests for translation.py
"""
def test_reverse_compliment():
    from tcrdist.translation import reverse_complement
    assert reverse_complement(seq = 'ATGC') == 'GCAT'
    assert reverse_complement(seq = 'AT.GC') == 'GC.AT'




import pytest



def test_example_11():
    """
    Example of comparing two clone_df from separate repertoires of different sizes
    
    1. Load a bulk beta-chain dataset with > 70 Clones.
    2. Prepare a beta chain clones_df DataFrame with cdr3_b_aa, pmhc_b_aa, cdr2_b_aa, cdr1_b_aa columns.
    3. Extract copy of processed dataframe.
    4. Load data of interest into a TCRrep instance.
    5. Compare the first 100 clones in primary dataset (100 Clones) against the the entire bulk dataset (77,538 Clones)
        Note: for large datasets you may want to turn store = False
    """
    import pandas as pd 
    from tcrdist.repertoire import TCRrep
    from tcrdist.setup_tests import download_and_extract_zip_file
    download_and_extract_zip_file('bulk.zip', source = "dropbox", dest = ".")
    df_bulk = pd.read_csv('bulk.csv', sep = ",")        #(1)
    tr_bulk = TCRrep(cell_df = df_bulk,                 #(2)
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = False)
    df_search = tr_bulk.clone_df.copy()                 #(3)

    df = pd.read_csv("dash.csv").head(100)              

    tr = TCRrep(cell_df = df,                           #(4)
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    
    assert tr.pw_beta.shape == (100,100)
    assert df_search.shape  == (77538, 8)
    tr.compute_rect_distances(df = tr.clone_df, df2 = df_search, store =True) #(5)
    
    assert tr.rw_beta.shape == (100, 77538)
    assert tr.rw_cdr3_b_aa.shape == (100, 77538)
    
    # see how many are tcrdistance less than 50
    import numpy as np
    np.sum(tr.rw_beta < 50, axis = 1)


"""
tcrdist3 hierahical clustering trees alpha/beta fill tests including
generating SVGs these tests take some time.
"""

def test_init_tree():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash.csv").sample(10).reset_index(drop = True)
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.mouse.ab.tree.html')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_size_olga_beta')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_unique_olga_beta')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('percent_missing_olga_beta')
    tcrtree.build_tree()

    assert os.path.isfile('dash.mouse.ab.tree.html')

def test_init_tree_beta():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash.csv").sample(10).reset_index(drop = True)
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.mouse.b.tree.html')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_size_olga_beta')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_unique_olga_beta')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('percent_missing_olga_beta')
    tcrtree.build_tree()
    assert os.path.isfile('dash.mouse.b.tree.html')

def test_init_tree_alpha():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash.csv").sample(10).reset_index(drop = True)
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.mouse.a.tree.html')



    tcrtree.build_tree()
    assert os.path.isfile('dash.mouse.a.tree.html')


def test_init_tree_human_beta():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash_human.csv").sample(10).reset_index(drop = True)
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.human.b.tree.html')
    tcrtree.combine_olga = True
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_size_olga_beta')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_unique_olga_beta')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('percent_missing_olga_beta')
    tcrtree.build_tree( )
    assert os.path.isfile('dash.human.b.tree.html')


def test_init_tree_human_alpha():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash_human.csv").sample(10).reset_index(drop = True)
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['alpha'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.human.a.tree.html')  
    tcrtree.combine_olga = True
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_size_olga_alpha')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('ref_unique_olga_alpha')
    tcrtree.default_plot_hclust_props['tooltip_cols'].append('percent_missing_olga_alpha')
    tcrtree.build_tree( )

    assert os.path.isfile('dash.human.a.tree.html')


def test_init_tree_human_alpha_beta():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash_human.csv").sample(10).reset_index(drop = True)
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['alpha', 'beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    tcrtree = TCRtree(tcrrep = tr, html_name = 'dash.human.ab.tree.html')
    tcrtree.build_tree()
    assert os.path.isfile('dash.human.ab.tree.html')





import pytest



def test_example_9():
    """
    If you already have a clones file and want 
    to compute 'tcrdistances' on a DataFrame with 
    custom columns names.
    
    Set:
    1. Assign TCRrep.clone_df
    2. set infer_cdrs = False,
    3. compute_distances = False
    4. deduplicate = False
    5. customize the keys for metrics, weights, and kargs with the lambda
        customize = lambda d : {new_cols[k]:v for k,v in d.items()} 
    6. call .calculate_distances()
    """
    import pwseqdist as pw
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    
    new_cols = {'cdr3_a_aa':'c3a', 'pmhc_a_aa':'pa', 'cdr2_a_aa':'c2a','cdr1_a_aa':'c1a',
                'cdr3_b_aa':'c3b', 'pmhc_b_aa':'pb', 'cdr2_b_aa':'c2b','cdr1_b_aa':'c1b'}
    
    df = pd.read_csv("dash2.csv").rename(columns = new_cols) 
    
    tr = TCRrep(
            cell_df = df,
            clone_df = df,              #(1)
            organism = 'mouse', 
            chains = ['alpha','beta'],
            infer_all_genes = True, 
            infer_cdrs = False,         #(2)s
            compute_distances = False,  #(3)
            deduplicate=False,          #(4)
            db_file = 'alphabeta_gammadelta_db.tsv')

    customize = lambda d : {new_cols[k]:v for k,v in d.items()} #(5)
    tr.metrics_a = customize(tr.metrics_a)
    tr.metrics_b = customize(tr.metrics_b)
    tr.weights_a = customize(tr.weights_a)
    tr.weights_b = customize(tr.weights_b)
    tr.kargs_a = customize(tr.kargs_a)
    tr.kargs_b = customize(tr.kargs_b)

    tr.compute_distances() #(6)
    
    # Notice that pairwise results now have custom names 
    tr.pw_c3b
    tr.pw_c3a
    tr.pw_alpha
    tr.pw_beta
    


"""
August 14, 2020
Unit tests for tcrdist.summarize
"""
import pytest
import numpy as np
import pandas as pd
from tcrdist.summarize import _summ, _occurs_N_str, _top_N_str

def test__top_N_str():
    df = pd.DataFrame({'catvar':["a","b","b","c"], "numvar":[10,1,100,3]})
    assert _top_N_str(m = df, col = 'catvar', count_col ='numvar', N=2) == 'b (88.6%), a (8.8%)'

def test__occurs_N_str():
    assert _occurs_N_str(["a","b","b","c"], 1) == 'b (50.0%)'    
    assert _occurs_N_str(["a","b","b","c"], 2) == 'b (50.0%), c (25.0%)' or _occurs_N_str(["a","b","b","c"], 2) == 'b (50.0%), a (25.0%)' 
    assert _occurs_N_str(["a","b","b","c"], 3) == 'b (50.0%), c (25.0%), a (25.0%)' or 'b (50.0%), a (25.0%), c (25.0%)'
 
def test__summ_0():
    """
    ValueError: No function (f) or function on a DataFrame (fdf) were supplied
    """
    df = pd.DataFrame({'catvar':["a","b","b","c"], "numvar":[10,1,100,3]})
    with pytest.raises(ValueError):
        r =  _summ(df, indices = [[0,1], [2,3]], column = 'numvar')

def test__summ_1():
    df = pd.DataFrame({'catvar':["a","b","b","c"], "numvar":[10,1,100,3]})
    r =  _summ(df, indices = [[0,1], [2,3]], column = 'numvar', f = np.median)
    e = [5.5, 51.5]
    assert np.all(r == e)

def test__summ_2():
    df = pd.DataFrame({'catvar':["a","b","b","b","b","c"], "numvar":[10,1,1,1,100,3]})
    r = _summ(df, indices = [[0,1,3], [3,4,5]], column = 'catvar', f = _occurs_N_str, N = 2)
    e = ['b (66.7%), a (33.3%)', 'b (66.7%), c (33.3%)']
    assert np.all(r == e)

def test__summ_3():
    df = pd.DataFrame({'catvar':["a","b","b","c"], "numvar":[10,1,100,3]})
    r = _summ(df, indices = [[0,1], [2,3]], column = 'catvar', fdf = _top_N_str, **{'col': 'catvar', 'count_col': 'numvar','N':2})
    e = ['a (90.9%), b (9.1%)', 'b (97.1%), c (2.9%)']
    assert np.all(r == e)




import pytest



def test_introduction_5():
    """
    Basic Specificity Neighborhoods based on a Fixed Radius
    """
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta','alpha'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    
    from tcrdist.rep_diff import neighborhood_diff
    # diff testing is pasted on binary comparison, so all epitope not 'PA' are set to 'X'
    tr.clone_df['PA'] = ['PA' if x == 'PA' else 'X' for x in tr.clone_df.epitope]
    
    # Larger Radius
    tr.nn_df = neighborhood_diff(clone_df= tr.clone_df, 
        pwmat = tr.pw_beta + tr.pw_alpha, 
        count_col = 'count', 
        x_cols = ['PA'], 
        knn_radius = 150)

    tr.nn_df[['K_neighbors', 'val_0', 'ct_0', 'val_2', 'ct_2', 'RR','OR', 'pvalue', 'FWERp','FDRq']].\
        sort_values(['FDRq']).sort_values(['OR','ct_0'], ascending = False)

    # Smaller Radius
    tr.nn_df = neighborhood_diff(clone_df= tr.clone_df, 
        pwmat = tr.pw_beta + tr.pw_alpha, 
        count_col = 'count', 
        x_cols = ['PA'], 
        knn_radius = 75)

    tr.nn_df[['K_neighbors', 'val_0', 'ct_0', 'val_2', 'ct_2', 'RR','OR', 'pvalue', 'FWERp','FDRq']].\
        sort_values(['FDRq']).sort_values(['OR','ct_0'], ascending = False)


import pytest
import pandas

def test_dash_ecdf():
    """
    An empirical distribution function (ECDF) can be created
    for a target TCR and a reference set of TCRs to show
    the proportion of reference TCRs that are within a distance
    D of the target TCR, over a range of distances.

    A plot of the ECDF as a function of increasing D shows the
    density of TCR space in the reference set in the neighborhood
    around the target TCR. This can be very helpful for 
    identifying dense antigen-specific clusters in an antigen
    enriched TCR repertoire, where the "reference" set is 
    actually an experimentally enriched repertoire (e.g. 
    pMHC:tetramer or AIM sorting). Or the ECDF can be helpful
    for identifying a radius around a TCR that retains high
    antigen specificity, by showing that the neighborhood
    is extremely sparse in an large unsorted/bulk TCR repertoire.
    
    """
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrsampler.sampler import TCRsampler
    from tcrdist.ecdf import distance_ecdf, make_ecdf_step
    from tcrdist.background import make_gene_usage_counter, make_vj_matched_background, \
                                    make_flat_vj_background, get_gene_frequencies, calculate_adjustment

    import matplotlib.pyplot as plt

    df = pd.read_csv('dash.csv')
    df = df.loc[df['epitope'] == 'PB1']
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    TCRsampler.download_background_file(download_file='wiraninha_sampler.zip')
    cols = ['v_b_gene','j_b_gene']

    refs = []
    for ts_fn in [f'wirasinha_mouse_beta_s_{i}.tsv.sampler.tsv' for i in '48']:
        ts = TCRsampler(default_background=ts_fn)
        ts.build_background(stratify_by_subject=True, use_frequency=False)
        
        """Sanitize the alleles to *01 for TCRSampler"""
        tmp = df[cols].applymap(lambda s: s.split('*')[0] + '*01')
        freqs = tmp.groupby(cols).size()
        freq_records = list(freqs.to_frame().to_records())
        ref = ts.sample(freq_records, depth=10, seed=110820)
        ref_df = pd.concat([pd.DataFrame({'cdr3_b_aa':ref[i]}).assign(v_b_gene=v, j_b_gene=j) for i,(v,j,_) in enumerate(freq_records)])        

        """Assigns pV, pJ and pVJ to ref_df"""
        ref_df = get_gene_frequencies(ts=ts, df=ref_df) 
        
        xdf = freqs.reset_index()
        xdf.columns = ['v_b_gene','j_b_gene', 'n']
        
        """For each V,J pairing compute frequency in this reference"""
        xdf = xdf.assign(ref_freq=xdf['n'] / xdf['n'].sum())
        ref_df = ref_df.merge(xdf, how='left', on=cols).reset_index()

        """ Assign weights to ref sequences: Pr_actual / Pr_sampling"""
        ref_df = ref_df.assign(weights=ref_df['pVJ'] / ref_df['ref_freq'])
        refs.append(ref_df)

        """Add uniformly sampled sequences"""
        ref_df = ts.ref_df.sample(100, random_state=1)
        refs.append(ref_df)

    ref_df = pd.concat(refs, axis=0)
    ref_tr = TCRrep(cell_df=ref_df[cols + ['cdr3_b_aa', 'weights']], 
                    organism='mouse', 
                    chains=['beta'],
                    compute_distances=False,
                    store_all_cdr=False)

    tr.compute_rect_distances(df=tr.clone_df, df2=ref_tr.clone_df, store=False)
    
    thresholds = np.arange(1, 50)
    thresholds, ref_ecdf = distance_ecdf(tr.rw_beta,
                                     thresholds=thresholds,
                                     weights=ref_tr.clone_df['weights'] * ref_tr.clone_df['count'])

    thresholds, target_ecdf = distance_ecdf(tr.pw_beta,
                                     thresholds=thresholds,
                                     weights=None)

    figh = plt.figure(figsize=(5, 5))
    axh = figh.add_axes([0.15, 0.15, 0.6, 0.7], yscale='log')
    plt.ylabel(f'Proportion of reference TCRs')
    plt.xlabel(f'Distance from target TCR clone')
    for tari in range(ref_ecdf.shape[0]):
        x, y = make_ecdf_step(thresholds, ref_ecdf[tari, :])
        axh.plot(x, y, color='k', alpha=0.2)
    x, y = make_ecdf_step(thresholds, np.mean(ref_ecdf, axis=0))
    axh.plot(x, y, color='r', alpha=1)

    figh = plt.figure(figsize=(5, 5))
    axh = figh.add_axes([0.15, 0.15, 0.6, 0.7], yscale='log')
    plt.ylabel(f'Proportion of target TCRs')
    plt.xlabel(f'Distance from target TCR clone')
    for tari in range(target_ecdf.shape[0]):
        x, y = make_ecdf_step(thresholds, target_ecdf[tari, :])
        axh.plot(x, y, color='k', alpha=0.2)
    x, y = make_ecdf_step(thresholds, np.mean(target_ecdf, axis=0))
    axh.plot(x, y, color='r', alpha=1)

    """Make an "ROC" plot combining the ECDF against the target (sensitivity)
    vs. ECDF against the reference (specificity)"""
    figh = plt.figure(figsize=(7, 5))
    axh = figh.add_axes([0.15, 0.15, 0.6, 0.7], yscale='log', xscale='log')
    plt.ylabel(f'Proportion of target TCRs')
    plt.xlabel(f'Proportion of reference TCRs')
    for tari in range(target_ecdf.shape[0]):
        x, y = make_ecdf_step(ref_ecdf[tari, :], target_ecdf[tari, :])
        axh.plot(x, y, color='k', alpha=0.2)
    x, y = make_ecdf_step(np.mean(ref_ecdf, axis=0), np.mean(target_ecdf, axis=0))
    axh.plot(x, y, color='r', alpha=1)
    yl = plt.ylim()
    xl = plt.xlim()
    #yl = (1e-6, 0.3)
    plt.plot(yl, yl, '--', color='gray')
    plt.xlim(xl)
    plt.ylim(yl)




import pytest 



def test_pgen_1():
    """
    How to add pgen estimates to human alpha/beta CDR3s
    """
    import pandas as pd
    from tcrdist.pgen import OlgaModel
    from tcrdist import mappers 
    from tcrdist.repertoire import TCRrep
    from tcrdist.setup_tests import download_and_extract_zip_file

    df = pd.read_csv("dash_human.csv")

    tr = TCRrep(cell_df = df.sample(5, random_state = 3), 
                organism = 'human', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv', 
                store_all_cdr = False)

    olga_beta  = OlgaModel(chain_folder = "human_T_beta", recomb_type="VDJ")
    olga_alpha = OlgaModel(chain_folder = "human_T_alpha", recomb_type="VJ")

    tr.clone_df['pgen_cdr3_b_aa'] = olga_beta.compute_aa_cdr3_pgens(
        CDR3_seq = tr.clone_df.cdr3_b_aa)
    
    tr.clone_df['pgen_cdr3_a_aa'] = olga_alpha.compute_aa_cdr3_pgens(
        CDR3_seq = tr.clone_df.cdr3_a_aa)

    tr.clone_df[['cdr3_b_aa', 'pgen_cdr3_b_aa', 'cdr3_a_aa','pgen_cdr3_a_aa']]
    """
                cdr3_b_aa  pgen_cdr3_b_aa          cdr3_a_aa  pgen_cdr3_a_aa
    0    CASSETSGRSPYEQYF    1.922623e-09    CAVRPGYSSASKIIF    1.028741e-06
    1  CASSPGLASPYSYNEQFF    1.554569e-11    CAVRVLMEYGNKLVF    1.128865e-08
    2       CASSSRSTDTQYF    5.184760e-07     CAGAGGGSQGNLIF    2.512580e-06
    3        CASSIGVYGYTF    8.576919e-08  CAFMSNAGGTSYGKLTF    1.832054e-07
    4  CASSLLVSGVSSTDTQYF    2.143508e-11       CAVIGEGGKLTF    5.698448e-12
    """


"""
An example for the gallery.
"""

def test_gallery_hdiff():
    """
    All imports are provided here, and are repeated 
    step-wise below, for clarity, and for
    module cut-and-paste. This example
    performs paired alpha-beta analysis,
    but code blocks can be used for single
    chain analysis as well.
    """
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_diff import hcluster_diff, member_summ
    from tcrsampler.sampler import TCRsampler
    from tcrdist.adpt_funcs import get_centroid_seq
    from tcrdist.summarize import _select
    from palmotif import compute_pal_motif, svg_logo
    from hierdiff import plot_hclust_props
    """
    Load a subset of data that contains paired alpha-beta
    chain mouse TCR receptors that recognized 
    the PA or PB1 epitopes (present in mouse influenza). 
    """
    import pandas as pd
    df = pd.read_csv("dash.csv")
    conditional = df['epitope'].apply( lambda x: x in ['PA','PB1'])
    """
    For illustrative/testing purposes, randomly subset the data to include 
    only 100 clones. Increase for more informative plot.
    """
    df = df[conditional].\
        reset_index(drop = True).\
        sample(100, random_state = 3).\
        reset_index(drop = True).\
        copy()
    """
    Load DataFrame into TCRrep instance, 
    which automatically computes attributes:
    1. .clone_df DataFrame
    2. .pw_beta nd.array 
    3. .pw_alpha nd.array 
    """
    from tcrdist.repertoire import TCRrep
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta','alpha'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    """
    Apply hcluster_diff, which hierarchically clusters.
    
    Note
    ----
    pwmat could easily be tr.pw_beta or tr.pw_alpha if 
    clustering should be done on a single chain.
    """
    from tcrdist.rep_diff import hcluster_diff
    tr.hcluster_df, tr.Z =\
        hcluster_diff(clone_df = tr.clone_df, 
                      pwmat    = tr.pw_beta + tr.pw_alpha,
                      x_cols = ['epitope'], 
                      count_col = 'count')

    """
    Load a custom background, mouse appropriate dataset to sample CDR3s 
    according to the V and J gene usage frequencies observed in each node.
    See the tcrsampler package for more details 
    (https://github.com/kmayerb/tcrsampler/blob/master/docs/getting_default_backgrounds.md)
    """
    from tcrsampler.sampler import TCRsampler

    t = TCRsampler()
    t.download_background_file("ruggiero_mouse_sampler.zip")
    tcrsampler_beta = TCRsampler(default_background = 'ruggiero_mouse_beta_t.tsv.sampler.tsv')
    tcrsampler_alpha = TCRsampler(default_background = 'ruggiero_mouse_alpha_t.tsv.sampler.tsv')

    """
    Add an SVG graphic to every node of the tree 
    aligned to the cluster centroid.
    """
    from tcrdist.adpt_funcs import get_centroid_seq
    from tcrdist.summarize import _select
    from palmotif import compute_pal_motif, svg_logo

    """Beta Chain"""
    svgs_beta = list()
    for i,r in tr.hcluster_df.iterrows():

        dfnode = tr.clone_df.iloc[r['neighbors_i'],]
        if dfnode.shape[0] > 2:
            centroid, *_ = get_centroid_seq(df = dfnode)
        else:
            centroid = dfnode['cdr3_b_aa'].to_list()[0]
        print(f"BETA-CHAIN: {centroid}")

        gene_usage_beta = dfnode.groupby(['v_b_gene','j_b_gene']).size()
        sampled_rep = tcrsampler_beta.sample( gene_usage_beta.reset_index().to_dict('split')['data'],
                        flatten = True, depth = 10)
        sampled_rep  = [x for x in sampled_rep if x is not None]
        motif, stat = compute_pal_motif(
                        seqs = _select(df = tr.clone_df, 
                                       iloc_rows = r['neighbors_i'], 
                                       col = 'cdr3_b_aa'),
                        refs = sampled_rep, 
                        centroid = centroid)
        
        svgs_beta.append(svg_logo(motif, return_str= True))

    """Add Beta SVG graphics to hcluster_df"""
    tr.hcluster_df['svg_beta'] = svgs_beta


    """Alpha Chain"""
    svgs_alpha = list()
    for i,r in tr.hcluster_df.iterrows():

        dfnode = tr.clone_df.iloc[r['neighbors_i'],]
        if dfnode.shape[0] > 2:
            centroid, *_ = get_centroid_seq(df = dfnode)
        else:
            centroid = dfnode['cdr3_a_aa'].to_list()[0]
        print(f"ALPHA-CHAIN: {centroid}")
        gene_usage_alpha = dfnode.groupby(['v_a_gene','j_a_gene']).size()
        sampled_rep = tcrsampler_alpha.sample( gene_usage_alpha.reset_index().to_dict('split')['data'], 
                        flatten = True, depth = 10)
        
        sampled_rep  = [x for x in sampled_rep if x is not None]
        motif, stat = compute_pal_motif(
                        seqs = _select(df = tr.clone_df, 
                                       iloc_rows = r['neighbors_i'], 
                                       col = 'cdr3_a_aa'),
                        refs = sampled_rep, 
                        centroid = centroid)

        svgs_alpha.append(svg_logo(motif, return_str= True))
    
    """Add Alpha SVG graphics to hcluster_df"""
    tr.hcluster_df['svg_alpha'] = svgs_alpha
    """
    Produce summary information for tooltips. 
    For instance, describe percentage of TCRs with 
    a given epitope at a given node.
    """
    res_summary = member_summ(  res_df = tr.hcluster_df,
                                clone_df = tr.clone_df, 
                                addl_cols=['epitope'])

    tr.hcluster_df_detailed = \
        pd.concat([tr.hcluster_df, res_summary], axis = 1)
    """
    Write D3 html for interactive denogram graphic. 
    Specify desired tooltips.
    """
    from hierdiff import plot_hclust_props
    html = plot_hclust_props(tr.Z,
                title='PA Epitope Example',
                res=tr.hcluster_df_detailed,
                tooltip_cols=['cdr3_b_aa','v_b_gene', 'j_b_gene','svg_alpha','svg_beta'],
                alpha=0.00001, colors = ['blue','gray'],
                alpha_col='pvalue')

    with open('hierdiff_example_PA_v_PB1.html', 'w') as fh:
        fh.write(html)